/** @file
 * new/delete邽߂̃ev[gB
 *
 * @author aki.
 */
#pragma once
#include "AllocatorImpl/Impl.h"
#include "AllocatorImpl/Log.h"
#include <stdlib.h>
#include <new>
#include <boost/static_assert.hpp>
namespace Themis {

namespace Detail {
    /// sstd::bad_alloc𓊂malloc()
    inline void* AllocatorMAlloc(size_t n) {
        void* p = malloc(n);
        if (p == NULL) throw std::bad_alloc();
        return p;
    }
    /// NULLĂdelete
    template<class T> inline void AllocatorSafeDelete(T*& rp) {
        T* p2 = rp;
        rp = NULL;
        delete p2;
    }
}

/**
    new/delete邽߂̃ev[gB

    Ƃ肠staticȔzpB
    fXgN^Ă΂ꂽthis͗Lȃ[wĂƂOł̎ȈׁA
    virtualƂ͎gȂB

    RsypTv(ǂ̃\[Xt@C)
@code

namespace {
 #if 0
    #define ALLOC_LOG   CAllocatorLog
 #else
    #define ALLOC_LOG   CAllocatorNoLog
 #endif
    typedef CAllocator<0x100, 0x100, ALLOC_LOG> TAllocator;
    TAllocator& GetAllocator() { static TAllocator o; return o; }
}
void* ::operator new(size_t n) {
    return GetAllocator().Alloc(n);
}
void ::operator delete(void* p) {
    GetAllocator().Free(p);
}

@endcode
*/
template<int S, int N, class L = CAllocatorNoLog> class CAllocator : private boost::noncopyable {
public:
    inline CAllocator() : mutex(NULL), m_oLogger(S, N) {
        mutex = new boost::mutex;
    }
    inline ~CAllocator() {
        Detail::AllocatorSafeDelete(mutex);
    }
    inline void* Alloc(size_t n) {
        Themis_DebugAssert(0 < n);
        if (mutex != NULL) {
            boost::mutex::scoped_lock lock(*mutex);
            return InnerAlloc(n);
        } else {
            // mutexȂrB
            return InnerAlloc(n);
        }
    }
    inline void Free(void* p) {
        if (p != NULL) {
            if (mutex != NULL) {
                boost::mutex::scoped_lock lock(*mutex);
                InnerFree(p);
            } else {
                // mutexȂrB
                InnerFree(p);
            }
        }
    }
private:
    inline void* InnerAlloc(size_t n) {
        if (m_oAllocator.IsAllocatable(n)) {
            m_oLogger.OnAllocBlock(n, m_oAllocator.GetBlankCount());
            return m_oAllocator.Allocate(n);
        } else {
            m_oLogger.OnAllocHeap(n);
            return Detail::AllocatorMAlloc(n);
        }
    }
    inline void InnerFree(void* p) {
        if (m_oAllocator.IsOwnMemory(p)) {
            m_oAllocator.Free(p);
        } else {
            free(p);
        }
        m_oLogger.OnFree();
    }
private:
    boost::mutex*           mutex;
    CAllocatorImpl<S, N>    m_oAllocator;
    L                       m_oLogger;
};

/**
    Ȃǂׂ̈ɂQAllocatorImplgŁB

    {IɃRsy{ςŁB
    S1 < S2ɂׂB

    S1 * N1 == S2 * N2 ƂƂɂāA
    N2 = N1 * S1 / S2 ƂĂ܂Ă̂A낤B̏Ԃ肵ȂBB
*/
template<int S1, int N1, int S2, int N2, class L = CAllocatorNoLog> class CAllocator2 : private boost::noncopyable {
public:
    BOOST_STATIC_ASSERT(S1 < S2);

    inline CAllocator2() : mutex(NULL), m_oLogger1(S1, N1), m_oLogger2(S2, N2) {
        mutex = new boost::mutex;
    }
    inline ~CAllocator2() {
        Detail::AllocatorSafeDelete(mutex);
    }
    inline void* Alloc(size_t n) {
        Themis_DebugAssert(0 < n);
        if (mutex != NULL) {
            boost::mutex::scoped_lock lock(*mutex);
            return InnerAlloc(n);
        } else {
            // mutexȂrB
            return InnerAlloc(n);
        }
    }
    inline void Free(void* p) {
        if (p != NULL) {
            if (mutex != NULL) {
                boost::mutex::scoped_lock lock(*mutex);
                InnerFree(p);
            } else {
                // mutexȂrB
                InnerFree(p);
            }
        }
    }
private:
    inline void* InnerAlloc(size_t n) {
        if (m_oAllocator1.IsAllocatable(n)) {
            m_oLogger1.OnAllocBlock(n, m_oAllocator1.GetBlankCount());
            m_oLogger2.OnAllocHeap(n);
            return m_oAllocator1.Allocate(n);
        } else if (m_oAllocator2.IsAllocatable(n)) {
            m_oLogger1.OnAllocHeap(n);
            m_oLogger2.OnAllocBlock(n, m_oAllocator2.GetBlankCount());
            return m_oAllocator2.Allocate(n);
        } else {
            m_oLogger1.OnAllocHeap(n);
            m_oLogger2.OnAllocHeap(n);
            return Detail::AllocatorMAlloc(n);
        }
    }
    inline void InnerFree(void* p) {
        if (m_oAllocator1.IsOwnMemory(p)) {
            m_oAllocator1.Free(p);
        } else if (m_oAllocator2.IsOwnMemory(p)) {
            m_oAllocator2.Free(p);
        } else {
            free(p);
        }
        m_oLogger1.OnFree();
        m_oLogger2.OnFree();
    }
private:
    boost::mutex*           mutex;
    CAllocatorImpl<S1, N1>  m_oAllocator1;
    CAllocatorImpl<S2, N2>  m_oAllocator2;
    L                       m_oLogger1;
    L                       m_oLogger2;
};

/**
    R̂B
    ܂ŗƏ_oȋC邯ǁBB
*/
template<int S1, int N1, int S2, int N2, int S3, int N3, class L = CAllocatorNoLog> class CAllocator3 : private boost::noncopyable {
public:
    BOOST_STATIC_ASSERT(S1 < S2);
    BOOST_STATIC_ASSERT(S2 < S3);

    inline CAllocator3() : mutex(NULL), m_oLogger1(S1, N1), m_oLogger2(S2, N2), m_oLogger3(S3, N3) {
        mutex = new boost::mutex;
    }
    inline ~CAllocator3() {
        Detail::AllocatorSafeDelete(mutex);
    }
    inline void* Alloc(size_t n) {
        Themis_DebugAssert(0 < n);
        if (mutex != NULL) {
            boost::mutex::scoped_lock lock(*mutex);
            return InnerAlloc(n);
        } else {
            // mutexȂrB
            return InnerAlloc(n);
        }
    }
    inline void Free(void* p) {
        if (p != NULL) {
            if (mutex != NULL) {
                boost::mutex::scoped_lock lock(*mutex);
                InnerFree(p);
            } else {
                // mutexȂrB
                InnerFree(p);
            }
        }
    }
private:
    inline void* InnerAlloc(size_t n) {
        if (m_oAllocator1.IsAllocatable(n)) {
            m_oLogger1.OnAllocBlock(n, m_oAllocator1.GetBlankCount());
            m_oLogger2.OnAllocHeap(n);
            m_oLogger3.OnAllocHeap(n);
            return m_oAllocator1.Allocate(n);
        } else if (m_oAllocator2.IsAllocatable(n)) {
            m_oLogger1.OnAllocHeap(n);
            m_oLogger2.OnAllocBlock(n, m_oAllocator2.GetBlankCount());
            m_oLogger3.OnAllocHeap(n);
            return m_oAllocator2.Allocate(n);
        } else if (m_oAllocator3.IsAllocatable(n)) {
            m_oLogger1.OnAllocHeap(n);
            m_oLogger2.OnAllocHeap(n);
            m_oLogger3.OnAllocBlock(n, m_oAllocator3.GetBlankCount());
            return m_oAllocator3.Allocate(n);
        } else {
            m_oLogger1.OnAllocHeap(n);
            m_oLogger2.OnAllocHeap(n);
            m_oLogger3.OnAllocHeap(n);
            return Detail::AllocatorMAlloc(n);
        }
    }
    inline void InnerFree(void* p) {
        if (m_oAllocator1.IsOwnMemory(p)) {
            m_oAllocator1.Free(p);
        } else if (m_oAllocator2.IsOwnMemory(p)) {
            m_oAllocator2.Free(p);
        } else if (m_oAllocator3.IsOwnMemory(p)) {
            m_oAllocator3.Free(p);
        } else {
            free(p);
        }
        m_oLogger1.OnFree();
        m_oLogger2.OnFree();
        m_oLogger3.OnFree();
    }
private:
    boost::mutex*           mutex;
    CAllocatorImpl<S1, N1>  m_oAllocator1;
    CAllocatorImpl<S2, N2>  m_oAllocator2;
    CAllocatorImpl<S3, N3>  m_oAllocator3;
    L                       m_oLogger1;
    L                       m_oLogger2;
    L                       m_oLogger3;
};


}
