#include "stdafx.h"
#include "SpiArchiver.h"
#include "ArchiverDll.h"
#include "../SPI.h"
#include "../FileBase.h"
#include "../Utility/StringUtility.h"
#include <time.h>
namespace Themis {

namespace {
    // SPI}`XbhɑΉĂƂ͌Ȃ̂ŁAÔߔrƂ
    boost::recursive_mutex& GetSpiArchiverExclusion() {
        static boost::recursive_mutex mutex;
        return mutex;
    }
}

CSpiArchiver::CSpiArchiver(const std::tr1::shared_ptr<CSpi>& pSpi): m_pSpi(pSpi) {
}

CSpiArchiver::~CSpiArchiver() {
}

vector<string> CSpiArchiver::InnerGetSupportedSuffixes() const {
    vector<string> vct;
    try {
        boost::recursive_mutex::scoped_lock lock(GetSpiArchiverExclusion());
        for (int i = 0; i < m_pSpi->GetCount(); i++) {
            if (!m_pSpi->IsArchiveExtractor(i)) {
                continue;   //ɗpȂ玟
            }
            try {
                SPI::GetPluginInfo gpi = m_pSpi->GetPluginInfo(i);
                Themis_DebugAssert(gpi != NULL);
                for (int i = 0; ; i++) {
                    char sz[0x80] = {0};
                    int n = gpi(i*2 + 2, sz, sizeof(sz) - 1);
                    if (n == 0) {
                        break;
                    }
                    sz[n] = L'\0';   //Ô߁Bn̒l͂܂AeɂȂȂǁBB
                    //  *.JPG Ƃ *.RGB;*.Q0 Ƃ炵
                    CSpi::SplitSuffixes(vct, Encoding::GetDefault().GetString(sz));
                }
            } catch (std::exception&) {
                // ֐|C^擾sH
            }
        }
    } catch (std::exception&) {
    }
    return vct;
}

bool CSpiArchiver::InnerOpen(vector<SFile>& rList) {
    try {
        auto_ptr<CApiFile> file;

        boost::recursive_mutex::scoped_lock lock(GetSpiArchiverExclusion());
        for (int i = 0; i < m_pSpi->GetCount(); i++) {
            if (!m_pSpi->IsArchiveExtractor(i)) {
                continue;   //ɗpȂ玟
            }
            try {
                SPI::IsSupported fnIsSupported = m_pSpi->IsSupported(i);
                SPI::GetArchiveInfo fnGetArchiveInfo = m_pSpi->GetArchiveInfo(i);
                SPI::GetFile fnGetFile = m_pSpi->GetFile(i);
                if (fnIsSupported == NULL || fnGetArchiveInfo == NULL || fnGetFile == NULL) {
                    Themis_DebugWarning(L"SPI̊֐擾Ɏs");
                    continue;
                }
                boost::scoped_array<char> sz(ToMultiByteChars(GetName()));

                if (file.get() == NULL) {
                    file.reset(new CApiFile(CFileName(GetName()), CApiFile::eModeOpenNoLock));
                } else {
                    try {
                        file->Seek(0, SEEK_SET);
                    } catch (std::exception&) {
                        // ɕĂƂŎXł(L`)
                        try { file.reset(); } catch (...) {}
                        file.reset(new CApiFile(CFileName(GetName()), CApiFile::eModeOpenNoLock));
                    }
                }
                Themis_DebugAssert(reinterpret_cast<uintptr_t>(file->GetApiHandle()) <= 0x0000ffff);

            BOOST_STATIC_ASSERT(sizeof(DWORD) == 4 && sizeof(HANDLE) == 4);
            #pragma warning(push)
            #pragma warning(disable : 4311)
                if (0 == fnIsSupported(sz.get(), reinterpret_cast<DWORD>(file->GetApiHandle()))) {
                    continue;   // vOCΉ
                }
            #pragma warning(pop)

             // try { file.reset(); } catch (...) {}

                HLOCAL hLocal = NULL;
                if (fnGetArchiveInfo(sz.get(), 0, 0, &hLocal) != 0 ||
                    ::LocalFlags(hLocal) == LMEM_INVALID_HANDLE) {
                    continue;
                }
                SPI::fileInfo* pInfo = static_cast<SPI::fileInfo*>(::LocalLock(hLocal));
                SPI::fileInfo* pInfo2 = pInfo;
                while (pInfo->method[0] != L'\0') {
                    if (pInfo->filename[0] != L'\0') {   //fBNg͖
                        SFile s;
                        s.strName = CFileName::AddDirSeparator(
                            Encoding::GetDefault().GetString(pInfo->path));
                        s.strName += Encoding::GetDefault().GetString(pInfo->filename);
                        s.nOrgSize = pInfo->filesize;

                        time_t tt = sizeof(time_t) <= 4 ?
                            static_cast<time_t>(pInfo->timestamp) :
                            static_cast<time_t>((time(NULL) & 0xffffffff00000000) | pInfo->timestamp);
                        
                        const tm* t = localtime(&tt);
                        SYSTEMTIME st = {0};
                        st.wYear = t->tm_year + 1900;
                        st.wMonth = t->tm_mon + 1;
                        st.wDayOfWeek = t->tm_wday;
                        st.wDay = t->tm_mday;
                        st.wHour = t->tm_hour;
                        st.wMinute = t->tm_min;
                        st.wSecond = t->tm_sec;
                     // st.wMilliseconds = 0;
                        BOOL bRet = ::SystemTimeToFileTime(&st, &s.ftLastUpdate);
                        Themis_DebugAssertApi(bRet != FALSE);
                        rList.push_back(s);
                        m_vctPositions.push_back(pInfo->position);
                    }
                    ++pInfo;
                }
                ::LocalUnlock(pInfo2);
                ::LocalFree(hLocal);
                m_nSpiIndex = i;
                return true; //
            } catch (...) {
                // sۂ̂Ŏ
            }
        }
        return false;   //throw RuntimeException(GetName() + L" ͖Ή̏ɂł");
    } catch (std::exception&) {
        return false;
    }
}

void CSpiArchiver::InnerExtractFile(int nFileIndex, const CFileName& oDirName) {
    boost::recursive_mutex::scoped_lock lock(GetSpiArchiverExclusion());
    try {
        LoadExtract(nFileIndex, oDirName);
        return;
    } catch (std::exception&) {
        Themis_DebugWarning(L"SPIł̓ǂݍݎsF" + GetName());
    }
    // SPIɏoB̕ꍇ͏ȂƎvǔÔ߁B
    SPI::GetFile fnGetFile = m_pSpi->GetFile(m_nSpiIndex);
    Themis_DebugAssert(fnGetFile != NULL);
    boost::scoped_array<char> szName(ToMultiByteChars(GetName()));
    boost::scoped_array<char> szDir(ToMultiByteChars(oDirName.RemoveDirSeparator()));
    int nError = fnGetFile(szName.get(), m_vctPositions[nFileIndex], szDir.get(), 0, NULL, 0);
    if (nError != 0) {
        throw RuntimeException(GetName() + L"  " + GetFileName(nFileIndex) + L" ̉𓀂Ɏs܂\r\nG[R[hF" + ItoA(nError));
    }
}

void CSpiArchiver::InnerLoadFile(int nFileIndex, vector<uint8_t>& vctData) {
    boost::recursive_mutex::scoped_lock lock(GetSpiArchiverExclusion());
    SPI::GetFile fnGetFile = m_pSpi->GetFile(m_nSpiIndex);
    Themis_DebugAssert(fnGetFile != NULL);
    HLOCAL hLocal = NULL;
    boost::scoped_array<char> szName(ToMultiByteChars(GetName()));
    int nError = fnGetFile(szName.get(), m_vctPositions[nFileIndex], reinterpret_cast<LPSTR>(&hLocal), 0x0100, NULL, 0);
    if (nError != 0 || ::LocalFlags(hLocal) == LMEM_INVALID_HANDLE) {
        throw RuntimeException(GetName() + L"  " + GetFileName(nFileIndex) + L" ̓ǂݍ݂Ɏs܂\r\nG[R[hF" + ItoA(nError));
    }
    Themis_DebugAssert(GetFileSize(nFileIndex) == ::LocalSize(hLocal));
    uint8_t* pByte = static_cast<uint8_t*>(::LocalLock(hLocal));
    vctData.assign(pByte, pByte + ::LocalSize(hLocal));
    ::LocalUnlock(pByte);
    ::LocalFree(hLocal);
}

std::tr1::shared_ptr<IArchiver> CSpiArchiverFactory::CreateInstance() {
    return std::tr1::shared_ptr<IArchiver>(new CSpiArchiver(m_pPluginList));
}

void CSpiArchiverFactory::SetDir(const CFileName& oSpiDir) {
    if (m_pPluginList == NULL) {
        m_pPluginList.reset(new CSpi(oSpiDir));
    } else {
        m_pPluginList->SetDir(oSpiDir);
    }
}


}
