#include "stdafx.h"
#include "Zip.h"
#if Themis_ConfigZLib
# include <boost/crc.hpp>
# include "../../../external/zlib.h"
# ifdef NDEBUG
#  pragma comment(lib, "zlib_mr.lib")
# else
#  pragma comment(lib, "zlib_md.lib")
# endif
# include "../TempFile.h"
# include "../File.h"
#endif // Themis_ConfigZLib
#include "ZipDecrypter.h"
#include "../String/Encoding.h"
#include "../Utility/Algorithm.h"
namespace Themis {

#if Themis_ConfigZLib

CZipArchiver::CZipArchiver() : m_bTryPass(false) {
}

CZipArchiver::~CZipArchiver() {
}

vector<string> CZipArchiver::InnerGetSupportedSuffixes() const {
    vector<string> vct;
    vct.push_back(L"zip");
    return vct;
}

bool CZipArchiver::InnerOpen(vector<SFile>& rList) {
    boost::mutex::scoped_lock lock(mutex);
    try {
     // m_pFile.reset(new CStdFile(CFileName(GetName()), L"rb"));
     // m_pFile.reset(new CApiFile(CFileName(GetName()), CApiFile::eModeOpen, CApiFile::eAccessSequential));
        m_pFile.reset(new CApiFile(CFileName(GetName()), CApiFile::eModeOpenNoLock));
        if (OpenByBackward(rList)) {
            return true;
        }
        Themis_DebugLog(L"CZipArchiver::OpenByBackward()s");
        rList.clear();
        if (OpenSequential(rList)) {
            Themis_DebugLog(L"CZipArchiver::OpenSequential()");
            return true;
        }
        return false;
    } catch (std::exception&) {
        m_pFile.reset();
        return false;
    }
}

bool CZipArchiver::OpenByBackward(vector<SFile>& rList) {
    try {
        m_bTryPass = false;
        // 64KBǂݍނ̂VۂǕʂɂȂɓǂ܂ȂĂ[ȁBB
        const long c_nEndRecMax = 0x10000;
        long nFileSize = long(m_pFile->GetSize());
        long nMemorySize = Min(c_nEndRecMax, nFileSize);
        boost::scoped_array<uint8_t> abyMemory(new uint8_t[nMemorySize]);
        // t@Cɓǂݍł܂
        m_pFile->Seek(nFileSize - nMemorySize, SEEK_SET);
        m_pFile->Read(abyMemory.get(), nMemorySize);
        Themis_DebugAssert(m_pFile->Tell() == long(m_pFile->GetSize()));
        // 납ǂ񂾕Ȃ̂Ō납T
        int nFileCount = -1;
        long nFileHeaderSize, nEndRecSize; // t@Cwb_̃TCYƃGhR[h̃TCY
        for (long i = nMemorySize - sizeof(Detail::SPK56); 0 <= i; i--) {
            Detail::SPK56* pk56 = reinterpret_cast<Detail::SPK56*>(&abyMemory[i]);
            if (pk56->dwSignature == 0x06054b50) { // wb_ꍇ
                nFileCount = pk56->wTotalDiskNumber;
                nFileHeaderSize = pk56->dwDirectorySize;
                nEndRecSize = nMemorySize - i;
                break;
            }
        }
        if (nFileCount < 0) {
            return false; // GhR[hȂ
        }
        // t@Cwb_p̊m
        long nMemoryHeaderPos;
        if (nFileHeaderSize + nEndRecSize < nMemorySize) { // ɓǂݍł镔ȏꍇ
            nMemoryHeaderPos = nMemorySize - (nFileHeaderSize + nEndRecSize);
        } else {
            if (nMemorySize < nFileHeaderSize) {
                nMemorySize = nFileHeaderSize;
                abyMemory.reset(new uint8_t[nMemorySize]); // mۂᑫ̂ōĊm
            }
            m_pFile->Seek(nFileSize - nMemorySize - nEndRecSize, SEEK_SET);
            m_pFile->Read(abyMemory.get(), nMemorySize);
            nMemoryHeaderPos = 0;
        }
        for (long i = 0, nPos = 0; i < nFileCount; i++) {
            if (nFileHeaderSize <= nPos) {
                return false; // Ȃ񂩕ςۂ
            }
            Detail::SPK12* pk12 = reinterpret_cast<Detail::SPK12*>(&abyMemory[nPos + nMemoryHeaderPos]);
            if (pk12->dwSignature != 0x02014b50) {
                return false; // Ȃ񂩕ςۂ
            }
            // TCY[̂Xg֒ǉ(TCYÔ͑SătH_ƌȂĖႤ)
            if (pk12->dwOriginalFileSize != 0) {
                SHeader oHeader;
                oHeader.nHeaderPos = pk12->dwLocalHeaderOffset;
                oHeader.nHeaderSize = -1; // ۂɓǂł݂񂱂Ƃɂ͕̂ŁBB
                if (pk12->wCompressionMethod == 0x0000) {
                 // wCompressionMethod == 0x0000 kۂǁAÔ߃TCY`FbNB
                 //  Íȏꍇ12Ă₱Ȃ̂ŃTCY`FbN͂߁B
                 // wExtractVersion0x000a H
                    oHeader.nCompressSize = 0; //k炵
                } else if (pk12->wCompressionMethod == 0x0008) { // deflateH
                    oHeader.nCompressSize = pk12->dwCompressSize;
                } else {
                    return false;
                }
                oHeader.bCrypted = (pk12->wGeneralFlags & 1) != 0;
                oHeader.cryptTest = (pk12->wGeneralFlags & 8) != 0 ? pk12->wUpdateTime : uint16_t(pk12->dwOriginalFileCrc >> 16);
                m_vctHeader.push_back(oHeader);
                char* p = reinterpret_cast<char*>(pk12 + 1);
                SFile s;
                s.strName = Encoding::GetDefault().GetString(std::string(p, p + pk12->wFileNameLength));
                s.nOrgSize = pk12->dwOriginalFileSize;
                BOOL bRet = ::DosDateTimeToFileTime(
                    pk12->wUpdateDate, pk12->wUpdateTime, &s.ftLastUpdate);
                Themis_DebugAssertApi(bRet != FALSE);
                rList.push_back(s);
            #ifndef NDEBUG
                Detail::SPK34 pk34;
                m_pFile->Seek(pk12->dwLocalHeaderOffset, SEEK_SET);
                m_pFile->Read(&pk34, sizeof(pk34));
                Themis_DebugAssert(pk34.dwSignature == 0x04034b50);
                Themis_DebugAssert(pk34.wCompressionMethod == pk12->wCompressionMethod);
                Themis_DebugAssert(pk34.dwOriginalFileSize == pk12->dwOriginalFileSize);
                Themis_DebugAssert(pk34.dwCompressSize == pk12->dwCompressSize);
                Themis_DebugAssert(pk34.wFileNameLength == pk12->wFileNameLength);
             // Themis_DebugAssert(pk34.wExtensionFieldLength == pk12->wExtensionFieldLength);
            #endif
            }
            // ̃wb_
            nPos += sizeof(Detail::SPK12) + pk12->wFileNameLength + pk12->wExtensionFieldLength + pk12->wCommentLength;
        }
        return true;
    } catch (std::exception&) {
        return false;
    }
}

bool CZipArchiver::OpenSequential(vector<SFile>& rList) {
    // ǂݍݑO
    m_vctHeader.clear();
    std::vector<char> vctFileName;
    m_pFile->Seek(0, SEEK_SET);
    // t@Cwb_̓ǂݍ
    while (true) {
        try {
            Detail::SPK34 pk34 = {0}; //ꉞOŖ߂Ƃ
            m_pFile->Read(&pk34, sizeof(pk34));
            if (pk34.dwSignature != 0x04034b50) {
                break;
            }
            // t@C
            if (0 < pk34.wFileNameLength) {
                vctFileName.clear();
                vctFileName.resize(pk34.wFileNameLength);
                m_pFile->Read(&vctFileName[0], vctFileName.size());
            } else {
                Themis_DebugWarning(L"ZIPɓɃt@C̖t@Cۂ");
                break; //wb_̓rŐK؂ȂRead()ŃRP͂ǁAÔ߁cB
            }
            // t@CʒuvZ
            size_t nHeaderPos = m_pFile->Tell() - sizeof(Detail::SPK34) - pk34.wFileNameLength;
            // g̈ƈkf[^΂Ď̂Ƃ܂ŁB
            size_t nFields = pk34.wExtensionFieldLength + pk34.dwCompressSize;
            if (0 < nFields) {
                m_pFile->Seek(static_cast<long>(nFields), SEEK_CUR);
            }
            // TCY[̂Xg֒ǉ(TCYÔ͑SătH_ƌȂĖႤ)
            if (pk34.dwOriginalFileSize != 0) {
                SHeader oHeader;
                oHeader.nHeaderPos = nHeaderPos;
                oHeader.nHeaderSize = sizeof(Detail::SPK34) + pk34.wFileNameLength + pk34.wExtensionFieldLength;
                if (pk34.wCompressionMethod == 0x0000) {
                 // wCompressionMethod == 0x0000 kۂǁAÔ߃TCY`FbNB
                 //  Íȏꍇ12Ă₱Ȃ̂ŃTCY`FbN͂߁B
                 // wExtractVersion0x000a H
                    oHeader.nCompressSize = 0; //k炵
                } else if (pk34.wCompressionMethod == 0x0008) { // deflateH
                    oHeader.nCompressSize = pk34.dwCompressSize;
                } else {
                    // Ȉk@̂Ȃ瑼ɓǂݍ߂̂ĂsƂƂɂ
                    m_vctHeader.clear();
                    break;
                }
                oHeader.bCrypted = (pk34.wGeneralFlags & 1) != 0;
                oHeader.cryptTest = (pk34.wGeneralFlags & 8) != 0 ? pk34.wUpdateTime : uint16_t(pk34.dwOriginalFileCrc >> 16);
                m_vctHeader.push_back(oHeader);

                SFile s;
                s.strName = Encoding::GetDefault().GetString(std::string(vctFileName.begin(), vctFileName.end()));
                s.nOrgSize = pk34.dwOriginalFileSize;
                BOOL bRet = ::DosDateTimeToFileTime(
                    pk34.wUpdateDate, pk34.wUpdateTime, &s.ftLastUpdate);
                Themis_DebugAssertApi(bRet != FALSE);
                rList.push_back(s);
            }
        } catch (std::exception&) {
            // K؂ꂾƂł̃G[ȏꍇɗ
            break;
        }
    }
    // wb_Ȃ΃G[ĎɂB(łǂݍ߂Ă珑ɂƂĈƂ)
    if (m_vctHeader.empty()) {
        m_pFile.reset();
        return false;
    }
    return true;
}

void CZipArchiver::InnerExtractFile(int nFileIndex, const CFileName& oDirName) {
    LoadExtract(nFileIndex, oDirName);
}

void CZipArchiver::InnerLoadFile(int nFileIndex, vector<uint8_t>& vctData) {
    boost::mutex::scoped_lock lock(mutex);
    // O
    vctData.clear();
    vctData.resize(GetFileSize(nFileIndex));
    if (vctData.empty()) return;

    if (m_vctHeader[nFileIndex].nHeaderSize != -1) {
        // wb_TCYmȏꍇ
        size_t n = m_vctHeader[nFileIndex].nHeaderPos + m_vctHeader[nFileIndex].nHeaderSize;
        m_pFile->Seek(static_cast<long>(n), SEEK_SET);
    } else {
        // wb_TCYmȏꍇ
        Detail::SPK34 pk34;
        m_pFile->Seek(static_cast<long>(m_vctHeader[nFileIndex].nHeaderPos), SEEK_SET);
        m_pFile->Read(&pk34, sizeof(pk34));
        m_vctHeader[nFileIndex].nHeaderSize = sizeof(Detail::SPK34) + pk34.wFileNameLength + pk34.wExtensionFieldLength;
        m_pFile->Seek(static_cast<long>(pk34.wFileNameLength + pk34.wExtensionFieldLength), SEEK_CUR);
    }
    // ÍĂꍇ
    auto_ptr<CZipDecrypter> pDecrypter;
    if (m_vctHeader[nFileIndex].bCrypted) {
        uint8_t by[12];
        m_pFile->Read(by, sizeof(by));
        if (m_bTryPass) {
            pDecrypter = CZipDecrypter::Create(m_pass, by, m_vctHeader[nFileIndex].cryptTest);
        } else {
            int tryCount = 0;
            string pass;
            while (QueryPassword(pass, tryCount++)) {
                pDecrypter = CZipDecrypter::Create(pass, by, m_vctHeader[nFileIndex].cryptTest);
                if (pDecrypter.get() != NULL) {
                    m_pass = pass;
                    break;
                }
            }
            m_bTryPass = true;
        }
        if (pDecrypter.get() == NULL) {
            throw RuntimeException(L"𓀂Ɏs܂");
        }
    }
    // 𓀏
    if (m_vctHeader[nFileIndex].nCompressSize == 0) {
        // k
        m_pFile->Read(&vctData[0], vctData.size());
        // Í
        if (pDecrypter.get() != NULL) {
            CZipDecrypter& r = *pDecrypter;
            for (size_t i = 0, n = vctData.size(); i < n; i++) {
                vctData[i] = r.decode(vctData[i]);
            }
        }
    } else {
        // k
        Inflate(&vctData[0], vctData.size(), nFileIndex, pDecrypter.get());
    }
}

void CZipArchiver::Inflate(uint8_t* p, size_t n, int nFileIndex, CZipDecrypter* pDecrypter) {
    uint8_t byBuffer[0x2000];
    size_t nCompressSize = m_vctHeader[nFileIndex].nCompressSize;
    if (pDecrypter != NULL) {
        nCompressSize -= 12; // cryptHeader
    }
    // zlibŉ𓀁[
    z_stream z = {0};
    z.zalloc = Z_NULL;
    z.zfree = Z_NULL;
    z.opaque = Z_NULL;
    z.next_out = p;
    z.avail_out = static_cast<uInt>(n);
    inflateInit2(&z, -MAX_WBITS);
    int nFlushFlag = Z_NO_FLUSH;
    while (true) {
        if (z.avail_in == 0) {
            size_t nRead = Min(nCompressSize, sizeof(byBuffer));
            if (0 < nRead) {
                m_pFile->Read(byBuffer, nRead);
                // Í
                if (pDecrypter != NULL) {
                    for (size_t i = 0; i < nRead; i++) {
                        byBuffer[i] = pDecrypter->decode(byBuffer[i]);
                    }
                }
            }
            z.next_in = byBuffer;
            z.avail_in = static_cast<uInt>(nRead);
            nCompressSize -= nRead;
            if (nCompressSize <= 0) nFlushFlag = Z_FINISH; // Zbginflateavail_outȂƃG[Ȃ̂Œ
        }
        int nRet = inflate(&z, nFlushFlag);
        if (nRet == Z_STREAM_END) {
            break;  //𓀂܂
        }
        if (nRet != Z_OK) {
            throw RuntimeException(GetName() + L"  " + GetFileName(nFileIndex) + L" ̉𓀂Ɏs܂");
        }
    }

    if (inflateEnd(&z) != Z_OK) {
        throw RuntimeException(GetName() + L"  " + GetFileName(nFileIndex) + L" ̉𓀌nɎs܂");
    }
}

//------------------------------------------------------------------//

CZipMaker::CZipMaker() {
}

CZipMaker::~CZipMaker() {
}

/// t@CEtH_̒ǉ
CZipMaker& CZipMaker::Add(const CFileName& oFileName) {
    if (CFile(oFileName).ExistsDir()) {
        InnerAddDir(oFileName.GetDirName(), oFileName);
    } else if (CFile(oFileName).ExistsFile()) {
        InnerAddDir(oFileName.GetPureName(), oFileName);
    } else {
        throw RuntimeException(oFileName.Get() + L" ݂܂B");
    }
    return *this;
}

/// t@CEtH_̒ǉ
CZipMaker& CZipMaker::Add(const string& strName, const CFileName& oFileName) {
    if (CFile(oFileName).ExistsDir()) {
        InnerAddDir(strName, oFileName);
    } else if (CFile(oFileName).ExistsFile()) {
        InnerAddFile(strName, oFileName);
    } else {
        throw RuntimeException(oFileName.Get() + L" ݂܂B");
    }
    return *this;
}

void CZipMaker::InnerAddFile(const string& strName, const CFileName& oFileName) {
    // t@CȃGgǉ
    m_vctEntry.push_back(SEntry(strName, oFileName, false));
}

void CZipMaker::InnerAddDir(const string& strName, const CFileName& oFileName) {
    Themis_DebugAssert(oFileName.IsFullPath());
    vector<CFileName> vct = CFile::Find(oFileName, 0);
    string strDirName = CFileName::RemoveDirSeparator(strName) + L"/";
    if (vct.empty()) {
        // tH_ȃGgǉ
        m_vctEntry.push_back(SEntry(strDirName, oFileName, true));
    } else {
        // tH_ȉ̃t@CċAIɏ
        for (size_t i = 0, n = vct.size(); i < n; i++) {
            string strRel = vct[i].GetRelativeName(oFileName, -1);
            Themis_DebugAssert(strRel.Substring(0, 3) != L"..\\");
            Add(strDirName + strRel, vct[i]);
        }
    }
}

CZipMaker::SEntry::SEntry(const string& s, const CFileName& f, bool d) : oFileName(f), bDir(d) {
    strName = s.Replace(L"\\", L"/");
}

/// ɂ̍쐬B
void CZipMaker::Save(const CFileName& oZipFileName) {
    CTempFile oDstTempFile(CFileName(oZipFileName.GetDir()), L"z", L"tmp");
    InnerSave(CStdFile(oDstTempFile.GetName(), L"wb"));
    oDstTempFile.MoveAndRelease(oZipFileName);
}

void CZipMaker::InnerSave(CStdFile& rDstFile) {
    CTempFile oTempFile(L"z", L"tmp");
    vector<Detail::SPK34> vctPK34List;
    vector<long> vctHeaderOffset;
    // PK34
    for (size_t i = 0, n = m_vctEntry.size(); i < n; i++) {
        Detail::SPK34 pk34 = {0};
        pk34.dwSignature = 0x04034b50;
    //
    //  wUpdateTime, wUpdateDateƂ悭ȂtOƂ͔΂Ă܂B
    //  KvɂȂ炻̂ƂׂB
    //
        std::string strName = Encoding::GetDefault().GetBytes(m_vctEntry[i].strName);
        pk34.wFileNameLength = static_cast<uint16_t>(strName.length());
        if (m_vctEntry[i].bDir) { // tH_ȏꍇ
            // wb_
            pk34.wExtractVersion = 0x000a;
         // pk34.wGeneralFlags
         // pk34.wCompressionMethod = 0x0008; // Z_DEFLATEDH
            ::FileTimeToDosDateTime(&CFile(m_vctEntry[i].oFileName).GetLastUpdate(), (LPWORD)&pk34.wUpdateDate, (LPWORD)&pk34.wUpdateTime);
         // pk34.dwOriginalFileCrc = dwCRC;
         // pk34.dwCompressSize = static_cast<uint32_t>(oSrcFile.GetSize());
         // pk34.dwOriginalFileSize = static_cast<uint32_t>(CFile(m_vctEntry[i].oFileName).GetSize());
            // ݁[
            vctHeaderOffset.push_back(rDstFile.Tell());
            rDstFile.Write(&pk34, sizeof(pk34));
            rDstFile.Write(strName.c_str(), strName.length());
        } else { // t@Cȏꍇ
            // k[
            CStdFile oSrcFile(oTempFile.GetName(), L"w+b");
            uint32_t dwCRC = Compress(oSrcFile, m_vctEntry[i].oFileName);
            // wb_
            pk34.wExtractVersion = 0x0014;
         // pk34.wGeneralFlags
            pk34.wCompressionMethod = 0x0008; // Z_DEFLATEDH
            ::FileTimeToDosDateTime(&CFile(m_vctEntry[i].oFileName).GetLastUpdate(), (LPWORD)&pk34.wUpdateDate, (LPWORD)&pk34.wUpdateTime);
            pk34.dwOriginalFileCrc = dwCRC;
            pk34.dwCompressSize = static_cast<uint32_t>(oSrcFile.GetSize());
            pk34.dwOriginalFileSize = static_cast<uint32_t>(CFile(m_vctEntry[i].oFileName).GetSize());
            // ݁[
            vctHeaderOffset.push_back(rDstFile.Tell());
            rDstFile.Write(&pk34, sizeof(pk34));
            rDstFile.Write(strName.c_str(), strName.length());
            vector<uint8_t> vct(pk34.dwCompressSize); // ʓ|Ȃ̂Ŏ蔲HB
            if (vct.size()) {
                oSrcFile.Seek(0, SEEK_SET);
                oSrcFile.Read(&vct[0], vct.size());
                rDstFile.Write(&vct[0], vct.size());
            }
        }
        // PK12pɃwb_XgɂĂ
        vctPK34List.push_back(pk34);
    }
    // PK12
    size_t nPK56DirSize = 0;
    for (size_t i = 0, n = m_vctEntry.size(); i < n; i++) {
        Detail::SPK12 pk12 = {0};
        pk12.dwSignature = 0x02014b50;
        pk12.wVersionMadeBy         = 0x0014;
        pk12.wExtractVersion        = vctPK34List[i].wExtractVersion;
        pk12.wGeneralFlags          = vctPK34List[i].wGeneralFlags;
        pk12.wCompressionMethod     = vctPK34List[i].wCompressionMethod;
        pk12.wUpdateTime            = vctPK34List[i].wUpdateTime;
        pk12.wUpdateDate            = vctPK34List[i].wUpdateDate;
        pk12.dwOriginalFileCrc      = vctPK34List[i].dwOriginalFileCrc;
        pk12.dwCompressSize         = vctPK34List[i].dwCompressSize;
        pk12.dwOriginalFileSize     = vctPK34List[i].dwOriginalFileSize;
        pk12.wFileNameLength        = vctPK34List[i].wFileNameLength;
        pk12.wExtensionFieldLength  = vctPK34List[i].wExtensionFieldLength;
     // pk12.wCommentLength
     // pk12.wDiskNumber
     // pk12.wInternalFileAttributes
     // pk12.dwExternalFileAttributes
        pk12.dwLocalHeaderOffset = vctHeaderOffset[i];
        // ݁[
        rDstFile.Write(&pk12, sizeof(pk12));
        std::string str = Encoding::GetDefault().GetBytes(m_vctEntry[i].strName);
        rDstFile.Write(str.c_str(), str.length());
        nPK56DirSize += sizeof(pk12) + str.length();
    }
    // PK56
    Detail::SPK56 pk56 = {0};
    pk56.dwSignature = 0x06054b50;
 // pk56.wDiskNumber
 // pk56.wDirectoryNumber
    pk56.wTotalDiskNumber = static_cast<uint16_t>(m_vctEntry.size());
    pk56.wTotalNumber = pk56.wTotalDiskNumber; // ̕ӂ悭̂BB
    pk56.dwDirectorySize = static_cast<uint32_t>(nPK56DirSize);
    pk56.dwOffsetDiskNum = static_cast<uint32_t>(rDstFile.Tell() - nPK56DirSize);
 // pk56.wCommentLength
    rDstFile.Write(&pk56, sizeof(pk56));
}

/// kCRCZo
uint32_t CZipMaker::Compress(CStdFile& rFile, const CFileName& oSrcFileName) {
    uint8_t bySrcBuf[0x1000];
    uint8_t byDstBuf[0x1000];
    // ̓t@CJ
    CStdFile oSrcFile(oSrcFileName, L"rb");
    size_t nSrcSize = oSrcFile.GetSize();
    oSrcFile.Seek(0, SEEK_SET);
    // CRCZo
    boost::crc_32_type crc;
    // zlibk
    z_stream z = {0};
    z.zalloc = Z_NULL;
    z.zfree = Z_NULL;
    z.opaque = Z_NULL;
    z.next_out = byDstBuf;
    z.avail_out = sizeof(byDstBuf);
    deflateInit2(&z, Z_DEFAULT_COMPRESSION, Z_DEFLATED, -MAX_WBITS, 8, Z_DEFAULT_STRATEGY);
        //Z_BEST_COMPRESSION, MAX_MEM_LEVELƂBB
    int nFlushFlag = Z_NO_FLUSH;
    while (true) {
        if (z.avail_in == 0) {
            // obt@ɓǂݍ
            if (sizeof(bySrcBuf) < nSrcSize) {
                oSrcFile.Read(bySrcBuf, sizeof(bySrcBuf));
                z.avail_in = static_cast<uInt>(sizeof(bySrcBuf));
                nSrcSize -= sizeof(bySrcBuf);
            } else {
                oSrcFile.Read(bySrcBuf, nSrcSize);
                z.avail_in = static_cast<uInt>(nSrcSize);
                nSrcSize = 0;
                nFlushFlag = Z_FINISH; // ܂B
            }
            z.next_in = bySrcBuf;
            // łCRCZo
            crc.process_bytes(z.next_in, z.avail_in);
        }
        int nRet = deflate(&z, nFlushFlag);
        if (nRet == Z_STREAM_END) {
            break;  //k܂
        }
        if (nRet != Z_OK) {
            throw RuntimeException(oSrcFileName.Get() + L" ̈kɎs܂");
        }
        // obt@Ȃ珑o
        if (z.avail_out == 0) {
            rFile.Write(byDstBuf, sizeof(byDstBuf));
            z.next_out = byDstBuf;
            z.avail_out = sizeof(byDstBuf);
        }
    }
    // co
    if (z.avail_out < sizeof(byDstBuf)) {
        size_t n = sizeof(byDstBuf) - z.avail_out;
        rFile.Write(byDstBuf, n);
    }
    // n
    if (deflateEnd(&z) != Z_OK) {
        throw RuntimeException(oSrcFileName.Get() + L" ̈knɎs܂");
    }
    // CRC̎Zo
    return crc.checksum();
}

#endif // Themis_ConfigZLib

}
