#include "stdafx.h"
#include "Calculator.h"
#include "CalculatorImpl/ScannerImpl.h"
#include "CalculatorImpl/FuncImpl.h"
#include "Utility/StringUtility.h"
namespace Themis {

CCalculator::CCalculator() {
    RegisterFunctions(m_oFuncList);
}

CCalculator::~CCalculator() {
}

bool CCalculator::IsDefinedVariable(const string& strName) {
    return InnerGetVariable(strName) != NULL;
}

CVariable& CCalculator::CreateVariable(const string& strName) {
    std::tr1::shared_ptr<CVariable> p(new CVariable(int(3735928559)));
    return *m_oVarList.insert(make_pair(strName, p)).first->second;
}

CVariable& CCalculator::GetVariable(const string& strName) {
    CVariable* pVar = InnerGetVariable(strName);
    if (pVar == NULL) {
        throw RuntimeException(L"ϐ '" + strName + L"' ܂");
    }
    return *pVar;
}

CVariable* CCalculator::InnerGetVariable(const string& strName) {
    TVarList::iterator it = m_oVarList.find(strName);
    return it != m_oVarList.end() ? it->second.get() : NULL;
}

void CCalculator::DeleteVariable(const string& strName) {
    m_oVarList.erase(strName);
}

void CCalculator::ClearVariables() {
    m_oVarList.clear();
}

void CCalculator::AddFunction(const string& strName, const std::tr1::shared_ptr<ICalculatorFunc>& pFunc) {
    m_oFuncList.insert(make_pair(strName, pFunc));
}

void CCalculator::RemoveFunction(const string& strName) {
    m_oFuncList.erase(strName);
}

void CCalculator::ClearFunctions() {
    m_oFuncList.clear();
}

CVariable CCalculator::Calc(const string& strExpression) {
    return Calc(CCalculatorScanner(strExpression, m_pErrorListener));
}

CVariableArray CCalculator::CalcCommaDivided(const string& strParams) {
    return CalcCommaDivided(CCalculatorScanner(strParams, m_pErrorListener));
}

std::tr1::shared_ptr<ICalculatorScanner> CCalculator::Prepare(const string& str) {
    return std::tr1::shared_ptr<ICalculatorScanner>(
        new CCompiledCalculatorScanner(CCalculatorScanner(str, m_pErrorListener)));
}

CVariable CCalculator::Calc(ICalculatorScanner& rScanner) {
    rScanner.Reset();
    CVariable v = Expression(rScanner);
    if (rScanner.GetToken() != eToken_EOF) {
        OnError(L"PExpressionȊO̕܂");
    }
    return v;
}

CVariableArray CCalculator::CalcCommaDivided(ICalculatorScanner& rScanner) {
    rScanner.Reset();
    return GetFunctionParams(rScanner);
}

//------------------------------------------------------------------//

CVariableArray CCalculator::GetFunctionParams(ICalculatorScanner& rScanner) {
    CVariableArray array;
    if (rScanner.GetToken() != eToken_EOF && rScanner.GetToken() != eToken_RPr) {
        array.Add(Expression(rScanner));
        while (rScanner.GetToken() == eToken_Comma) {
            rScanner.ToNext();
            array.Add(Expression(rScanner));
        }
    }
    return array;
}

CVariable CCalculator::Expression(ICalculatorScanner& rScanner) {
    CVariable x = Expression2(rScanner);
    while (true) {
        if (rScanner.GetToken() == eToken_Question) {
            rScanner.ToNext();
            CVariable y = Expression(rScanner);
            if (rScanner.GetToken() == eToken_Colon) {
                rScanner.ToNext();
                CVariable z = Expression(rScanner);
                x = x.GetAsBool()? y: z;
            } else {
                OnError(L"Zq ? ̋؂蕶 : ܂");
            }
        } else {
            break;  //YTokenȂȂI
        }
    }
    return x;
}

CVariable CCalculator::Expression2(ICalculatorScanner& rScanner) {
    CVariable x = Expression3(rScanner);
    while (true) {
        if (rScanner.GetToken() == eToken_And2) {
            rScanner.ToNext();
            CVariable y = Expression3(rScanner);
            x = x.GetAsBool() && y.GetAsBool();
        } else if (rScanner.GetToken() == eToken_Or2) {
            rScanner.ToNext();
            CVariable y = Expression3(rScanner);
            x = x.GetAsBool() && y.GetAsBool();
        } else {
            break;  //YTokenȂȂI
        }
    }
    return x;
}

CVariable CCalculator::Expression3(ICalculatorScanner& rScanner) {
    CVariable x = Term(rScanner);
    while (true) {
        if (rScanner.GetToken() == eToken_Gt) {
            rScanner.ToNext();
            CVariable y = Term(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() > y.GetInt(); break;
            case 1: x = x.GetFloat() > y.GetFloat(); break;
            case 2:
                if (x.IsNum() || y.IsNum()) {
                    OnError(L"Ɛl̔r͍s܂");
                } else {
                    x = x.GetStr() > y.GetStr();
                }
                break;
            default: Themis_DebugTrespass();
            }
        } else if (rScanner.GetToken() == eToken_Ge) {
            rScanner.ToNext();
            CVariable y = Term(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() >= y.GetInt(); break;
            case 1: x = x.GetFloat() >= y.GetFloat(); break;
            case 2:
                if (x.IsNum() || y.IsNum()) {
                    OnError(L"Ɛl̔r͍s܂");
                } else {
                    x = x.GetStr() >= y.GetStr();
                }
                break;
            default: Themis_DebugTrespass();
            }
        } else if (rScanner.GetToken() == eToken_Lt) {
            rScanner.ToNext();
            CVariable y = Term(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() < y.GetInt(); break;
            case 1: x = x.GetFloat() < y.GetFloat(); break;
            case 2:
                if (x.IsNum() || y.IsNum()) {
                    OnError(L"Ɛl̔r͍s܂");
                } else {
                    x = x.GetStr() < y.GetStr();
                }
                break;
            default: Themis_DebugTrespass();
            }
        } else if (rScanner.GetToken() == eToken_Le) {
            rScanner.ToNext();
            CVariable y = Term(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() <= y.GetInt(); break;
            case 1: x = x.GetFloat() <= y.GetFloat(); break;
            case 2:
                if (x.IsNum() || y.IsNum()) {
                    OnError(L"Ɛl̔r͍s܂");
                } else {
                    x = x.GetStr() <= y.GetStr();
                }
                break;
            default: Themis_DebugTrespass();
            }
        } else if (rScanner.GetToken() == eToken_Eq) {
            rScanner.ToNext();
            CVariable y = Term(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() == y.GetInt(); break;
            case 1: x = x.GetFloat() == y.GetFloat(); break;
            case 2:
                if (x.IsNum() || y.IsNum()) {
                    OnError(L"Ɛl̔r͍s܂");
                } else {
                    x = x.GetStr() == y.GetStr();
                }
                break;
            default: Themis_DebugTrespass();
            }
        } else if (rScanner.GetToken() == eToken_Ne) {
            rScanner.ToNext();
            CVariable y = Term(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() != y.GetInt(); break;
            case 1: x = x.GetFloat() != y.GetFloat(); break;
            case 2:
                if (x.IsNum() || y.IsNum()) {
                    OnError(L"Ɛl̔r͍s܂");
                } else {
                    x = x.GetStr() != y.GetStr();
                }
                break;
            default: Themis_DebugTrespass();
            }
        } else {
            break;  //YTokenȂȂI
        }
    }
    return x;
}

CVariable CCalculator::Term(ICalculatorScanner& rScanner) {
    CVariable x = Term2(rScanner);
    while (true) {
        if (rScanner.GetToken() == eToken_Plus) {
            rScanner.ToNext();
            CVariable y = Term2(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() + y.GetInt(); break;
            case 1: x = x.GetFloat() + y.GetFloat(); break;
            case 2:
                if (x.IsNum() || y.IsNum()) {
                    OnError(L"Ɛl̉Z͍s܂");
                } else {
                    x = x.GetStr() + y.GetStr();
                }
                break;
            default: Themis_DebugTrespass();
            }
        } else if (rScanner.GetToken() == eToken_Minus) {
            rScanner.ToNext();
            CVariable y = Term2(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() - y.GetInt(); break;
            case 1: x = x.GetFloat() - y.GetFloat(); break;
            case 2: OnError(L"ɌZ͍s܂"); break;
            default: Themis_DebugTrespass();
            }
        } else if (rScanner.GetToken() == eToken_ShR) {
            rScanner.ToNext();
            CVariable y = Term2(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() >> y.GetInt(); break;
            case 1: OnError(L"lɉEVtgZ͍s܂"); break;
            case 2: OnError(L"ɉEVtgZ͍s܂"); break;
            default: Themis_DebugTrespass();
            }
        } else if (rScanner.GetToken() == eToken_ShL) {
            rScanner.ToNext();
            CVariable y = Term2(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() << y.GetInt(); break;
            case 1: OnError(L"lɍVtgZ͍s܂"); break;
            case 2: OnError(L"ɍVtgZ͍s܂"); break;
            default: Themis_DebugTrespass();
            }
        } else if (rScanner.GetToken() == eToken_And) {
            rScanner.ToNext();
            CVariable y = Term2(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() & y.GetInt(); break;
            case 1: OnError(L"lɘ_ωZ & ͍s܂"); break;
            case 2: OnError(L"ɘ_ωZ & ͍s܂"); break;
            default: Themis_DebugTrespass();
            }
        } else if (rScanner.GetToken() == eToken_Or) {
            rScanner.ToNext();
            CVariable y = Term2(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() | y.GetInt(); break;
            case 1: OnError(L"lɘ_aZ | ͍s܂"); break;
            case 2: OnError(L"ɘ_aZ | ͍s܂"); break;
            default: Themis_DebugTrespass();
            }
        } else if (rScanner.GetToken() == eToken_Xor) {
            rScanner.ToNext();
            CVariable y = Term2(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() ^ y.GetInt(); break;
            case 1: OnError(L"lɔrI_aZ ^ ͍s܂"); break;
            case 2: OnError(L"ɔrI_aZ ^ ͍s܂"); break;
            default: Themis_DebugTrespass();
            }
        } else {
            break;  //YTokenȂȂI
        }
    }
    return x;
}

CVariable CCalculator::Term2(ICalculatorScanner& rScanner) {
    CVariable x = Factor(rScanner);
    while (true) {
        if (rScanner.GetToken() == eToken_Mul) {
            rScanner.ToNext();
            CVariable y = Factor(rScanner);
            switch (Max(x.GetType(), y.GetType())) {
            case 0: x = x.GetInt() * y.GetInt(); break;
            case 1: x = x.GetFloat() * y.GetFloat(); break;
            case 2: OnError(L"ɏZ͍s܂"); break;
            default: Themis_DebugTrespass();
            }
        } else if (rScanner.GetToken() == eToken_Div) {
            rScanner.ToNext();
            CVariable y = Factor(rScanner);
            int iy;
            double dy;
            switch (Max(x.GetType(), y.GetType())) {
            case 0:
                iy = y.GetInt();
                if (iy == 0) {
                    OnError(L"̏Zŕꂪ0ł");
                    //Ƃ肠y+0Ɖ肵ċɌlݒ肵Ă
                    x = 0 <= x.GetInt()? INT_MAX: INT_MIN;
                } else {
                    x = x.GetInt() / iy;
                }
                break;
            case 1:
                dy = y.GetFloat();
                if (dy == 0.0) {
                    OnError(L"̏Zŕꂪ0ł");
                    //Ƃ肠y𐳂̒l̉Ɖ肵ċɌlݒ肵Ă
                    x = 0.0 <= x.GetFloat()? DBL_MAX: DBL_MIN;
                } else {
                    x = x.GetFloat() / dy;
                }
                break;
            case 2: OnError(L"ɏZ͍s܂"); break;
            default: Themis_DebugTrespass();
            }
        } else if (rScanner.GetToken() == eToken_Mod) {
            rScanner.ToNext();
            CVariable y = Factor(rScanner);
            int iy;
            switch (Max(x.GetType(), y.GetType())) {
            case 0:
                iy = y.GetInt();
                if (iy == 0) {
                    OnError(L"̏]ZŊ鐔0ł");
                    //Ƃ肠y+0Ɖ肵ċɌlݒ肵Ă
                    x = 0 <= x.GetInt()? INT_MAX: INT_MIN;
                } else {
                    x = x.GetInt() % y.GetInt();
                }
                break;
            case 1:
                x = fmod(x.GetFloat(), y.GetFloat());
                break;
            case 2: OnError(L"ɏ]vZ͍s܂"); break;
            default: Themis_DebugTrespass();
            }
        } else {
            break;  //YTokenȂȂI
        }
    }
    return x;
}

CVariable CCalculator::Factor(ICalculatorScanner& rScanner) {
    if (rScanner.GetToken() == eToken_Num) {                //l
        int n = rScanner.GetNum();
        rScanner.ToNext();
        return n;
    } else if (rScanner.GetToken() == eToken_NumF) {          //l
        double d = rScanner.GetNumF();
        rScanner.ToNext();
        return d;
    } else if (rScanner.GetToken() == eToken_Str) {           //񃊃e
        string str = rScanner.GetStr();
        rScanner.ToNext();
        return str;
    } else if (rScanner.GetToken() == eToken_Minus) {         //P}CiX
        rScanner.ToNext();
        CVariable x = Factor(rScanner);
        switch (x.GetType()) {
        case 0: return 0 - x.GetInt();
        case 1: return 0.0 - x.GetFloat();
        case 2: OnError(L"P}CiX̌͐l̂łȂĂ͂Ȃ܂");
                return CVariable(0);   //Ƃ肠_~[̒lB
        default: Themis_DebugTrespass();
        }
    } else if (rScanner.GetToken() == eToken_Plus) {          //PvX
        rScanner.ToNext();
        return Factor(rScanner);    //zgGetType()ĒׂKv̂ǂc
    } else if (rScanner.GetToken() == eToken_CPL) {           //␔
        rScanner.ToNext();
        CVariable x = Factor(rScanner);
        if (x.IsNum()) {
            return ~ x.GetInt();
        } else {
            OnError(L"␔Zq̌͐l̂łȂĂ͂Ȃ܂");
            return CVariable(0);   //Ƃ肠_~[̒lB
        }
    } else if (rScanner.GetToken() == eToken_Not) {           //_ے
        rScanner.ToNext();
        CVariable x = Factor(rScanner);
        return !x.GetAsBool();
    } else if (rScanner.GetToken() == eToken_LPr) {           //iexpressionj
        rScanner.ToNext();
        CVariable x = Expression(rScanner);
        if (rScanner.GetToken() == eToken_RPr) {
            rScanner.ToNext();
        } else {
            OnError(L" ) ܂");
        }
        return x;
    } else if (rScanner.GetToken() == eToken_Label) {         //x(֐ĂяoEϐQ)
        string strLabel = rScanner.GetStr();
        rScanner.ToNext();
        if (rScanner.GetToken() == eToken_LPr) {                //x (
            rScanner.ToNext();
            //֐ƌȂ
            CVariableArray oParams = GetFunctionParams(rScanner);
            if (rScanner.GetToken() != eToken_RPr) {            //̕
                OnError(L"֐̈̕ ) ܂");
            }
            rScanner.ToNext();
            CVariable oRet;
            if (CallFunction(strLabel, oRet, oParams)) {
                return oRet;
            } else {
                OnError(L"`̊֐ '" + strLabel + L"' Ăяo܂");
                return CVariable(0);
            }
        } else {
            //ϐƌȂ
            string strSelector = Selector(rScanner);
            CVariable* p = InnerGetVariable(strLabel + strSelector);
            if (p != NULL) {
                return *p;
            } else {
                OnError(L"ϐ '" + strLabel + strSelector + L"' `Ă܂B0(l)ƌȂ܂");
                return CVariable(0);
            }
        }
    }
    OnError(L"l\vf܂");
    return CVariable(0);
}

string CCalculator::Selector(ICalculatorScanner& rScanner) {
    string strSelector;
    while (true) {
        if (rScanner.GetToken() == eToken_Dot) {
            //ƂAo֐悤ƂƂŉ͂̂̓CJCcB
            //x̎ʂȂ烁o֐AłȂ΃oϐĎɂȂ̂ǁA
            // !GetWindow().IsVisible() ݂Ɋ֐̖߂lɂĂĂPZqDxA
            //Factor̒ōlS̏ꏊSelector̂H
            rScanner.ToNext();
            if (rScanner.GetToken() == eToken_Label) {
                strSelector += string(L'.', 1);
                strSelector += rScanner.GetStr();
                rScanner.ToNext();
            } else {
                OnError(L"oQƉZq . ̌Ƀx܂");
            }
        } else if (rScanner.GetToken() == eToken_LSq) {
            CVariable x = Expression(rScanner);
            if (rScanner.GetToken() == eToken_RSq) {
                if (x.IsNum()) {
                    strSelector += string::Format(L":%d", x.GetInt());
                } else {
                    OnError(L"z̓Yɕ͎gps\ł");
                }
                rScanner.ToNext();
            } else {
                OnError(L"zYw̕ ] ܂");
            }
        } else {
            return strSelector;
        }
    }
}

//------------------------------------------------------------------//

void CCalculator::OnError(const string& strMsg) const {
    Themis_DebugLog(strMsg); // Warning̕H 납B
    if (m_pErrorListener.get() != NULL) {
        m_pErrorListener->OnError(strMsg);
    }
}

bool CCalculator::CallFunction(const string& strFunc, CVariable& rRet, const CVariableArray& rParams) {
    TFuncList::iterator it = m_oFuncList.find(strFunc);
    if (it != m_oFuncList.end()) {
        return it->second->Exec(*this, strFunc, rRet, rParams);
    }
    return false;
}

//------------------------------------------------------------------//

string CCalculator::Replace2Byte(const string& s) {
    vector<pair<string, string> > vct;
    vct.push_back(make_pair(L"{", L"+"));
    vct.push_back(make_pair(L"|", L"-"));
    vct.push_back(make_pair(L"", L"*"));
    vct.push_back(make_pair(L"^", L"/"));
    vct.push_back(make_pair(L"~", L"*"));
    vct.push_back(make_pair(L"", L"/"));
    vct.push_back(make_pair(L"i", L"("));
    vct.push_back(make_pair(L"j", L")"));
    vct.push_back(make_pair(L"[", L"("));
    vct.push_back(make_pair(L"]", L")"));
    vct.push_back(make_pair(L"{", L"("));
    vct.push_back(make_pair(L"}", L")"));
    vct.push_back(make_pair(L"m", L"("));
    vct.push_back(make_pair(L"n", L")"));
    vct.push_back(make_pair(L"o", L"("));
    vct.push_back(make_pair(L"p", L")"));
    vct.push_back(make_pair(L"O", L"0"));
    vct.push_back(make_pair(L"P", L"1"));
    vct.push_back(make_pair(L"Q", L"2"));
    vct.push_back(make_pair(L"R", L"3"));
    vct.push_back(make_pair(L"S", L"4"));
    vct.push_back(make_pair(L"T", L"5"));
    vct.push_back(make_pair(L"U", L"6"));
    vct.push_back(make_pair(L"V", L"7"));
    vct.push_back(make_pair(L"W", L"8"));
    vct.push_back(make_pair(L"X", L"9"));
    vct.push_back(make_pair(L"D", L"."));
    vct.push_back(make_pair(L"", L"pi"));
    return Replace(s, vct);
}

}
