#include "stdafx.h"
#include "Csv.h"
#include "File.h"
#include "Char.h"
#include "StringScanner.h"
#include "String/StringBuilder.h"
#include "String/Encoding.h"
namespace Themis {

CCsv::CCsv() {
}

CCsv::CCsv(const CFileName& oFileName) {
    Load(oFileName);
}

CCsv::~CCsv() {
}

void CCsv::Load(const CFileName& oFileName) {
    LoadFromStr(CFile(oFileName).LoadStr());
}

void CCsv::Save(const CFileName& oFileName) {
    CFile(oFileName).Save(SaveToStr());
}

void CCsv::LoadFromStr(const string& strData) {
    m_vctData.clear();
    vector<string> vctTmpLine;
    StringBuilder tmpData;
    bool bWQuote = false;
    CStringScanner oScanner(strData);
    while (oScanner) {
        if (bWQuote) { // L""̒
            if (oScanner.IsMatch(L'"', L'"')) {
                tmpData.Append(L"\"");
            } else if (oScanner.IsMatch(L'"')) {
                bWQuote = false;
            } else {
                oScanner.Copy1(tmpData);
            }
        } else { // L""̊O
            if (tmpData.GetLength() <= 0 && oScanner.IsMatch(L'"')) {
                bWQuote = true;
            } else if (oScanner.IsMatch(L',')) { // J}
                vctTmpLine.push_back(tmpData.ToString());
                tmpData.Remove(0, tmpData.GetLength());
            } else if (oScanner.IsMatchNewline()) { // s
                m_vctData.push_back(vctTmpLine);
                vctTmpLine.clear();
            } else {
                oScanner.Copy1(tmpData);
            }
        }
    }
}

namespace {
    inline string CsvEscape(const string& str) {
        // J}sĂ邩ǂ`FbNB
        if (0 <= str.IndexOf(L',') || 0 <= str.IndexOf(L'\r') || 0 <= str.IndexOf(L'\n')) {
            return L"\"" + str.Replace(L"\"", L"\"\"") + L"\"";
        }
        return str;
    }
}
string CCsv::SaveToStr() {
    StringBuilder builder;
    for (int i = 0, nLines = GetLineCount(); i < nLines; i++) {
        for (int k = 0, nRows = GetRowCount(i); k < nRows; k++) {
            if (k != 0) {
                builder.Append(L',');
            }
            builder.Append(CsvEscape(Get(i, k)));
        }
        builder.AppendLine();
    }
    return builder.ToString();
}

//------------------------------------------------------------------//

}
#include <fstream>
namespace Themis {

struct SCsvReaderData {
    std::ifstream ifs;
    unsigned nFileSize;
};

CCsvReader::CCsvReader(const CFileName& oFileName) {
/*
    data->nFileSize = static_cast<unsigned>(CFile(oFileName).GetSize());
    data->ifs.open(Encoding::GetDefault().GetBytes(oFileName.Get()).c_str());
/*/
    data->ifs.open(Encoding::GetDefault().GetBytes(oFileName.Get()).c_str());
    data->ifs.seekg(0, std::ios::end);
    data->nFileSize = data->ifs.tellg();
    data->ifs.seekg(0, std::ios::beg);
//*/
}

CCsvReader::~CCsvReader() {
}

/// ̍sֈړBI[ɓBĂfalse
bool CCsvReader::ToNextLine() {
    std::string line;
    if (std::getline(data->ifs, line)) {
        m_strOldLine = Encoding::GetDefault().GetString(line);
        m_vctCols = m_strOldLine.Split(Array(L','));
        for (size_t i = 0, n = m_vctCols.size(); i < n; i++) {
            if (2 <= m_vctCols[i].GetLength() &&
                m_vctCols[i].StartsWith(L"\"") && 
                m_vctCols[i].EndsWith(L"\"")) {
                m_vctCols[i] = m_vctCols[i].Substring(1, m_vctCols[i].GetLength() - 2);
            }
        }
        return !m_vctCols.empty();
    }
    return false;
}

//@{ i\po
unsigned CCsvReader::GetCurrentPos() const {
    return data->ifs.tellg();
}
unsigned CCsvReader::GetSize() const {
    return data->nFileSize;
}
//@}

}
