#pragma once
namespace Themis {

template<class T> class ComponentFactory0 : IFactory<void> {
    virtual std::tr1::shared_ptr<void> CreateInstance() { return new T(); }
};

template<class T, class A> class ComponentFactory1 : IFactory<void> {
    A a;
    ComponentFactory1(typedef typename boost::call_traits<A>::param_type a) : a(a) {}
    virtual std::tr1::shared_ptr<void> CreateInstance() { return new T(a); }
};

template<class T, class A, class B> class ComponentFactory2 : IFactory<void> {
    A a;
    B b;
    ComponentFactory1(typedef typename boost::call_traits<A>::param_type a,
        typedef typename boost::call_traits<B>::param_type b) : a(a), b(b) {}
    virtual std::tr1::shared_ptr<void> CreateInstance() { return new T(a, b); }
};

template<class T, class A, class B, class C> class ComponentFactory3 : IFactory<void> {
    A a;
    B b;
    C c;
    ComponentFactory1(typedef typename boost::call_traits<A>::param_type a,
        typedef typename boost::call_traits<B>::param_type b,
        typedef typename boost::call_traits<C>::param_type c) : a(a), b(b), c(c) {}
    virtual std::tr1::shared_ptr<void> CreateInstance() { return new T(a, b, c); }
};

/**
 * DependencyInjectionp̃v[Xz_B
 */
template<class T> class Component {}

/**
 * Component̃p[^B
 */
struct ComponentDef {
    std::tr1::shared_ptr<IFactory<void> > factory;
    // setters
};

/**
 * DIContainerۂB
 */
class DIContainer {
    std::map<string, ComponentDef> componentDefs;
    std::map<string, std::tr1::shared_ptr<void> > instances;
public:
    /// ComponentDef̃ZbgB
    void Set(const string& name, const ComponentDef& componentDef) {
        componentDefs.insert(name, componentDef);
    }

    /// IuWFNg𒼐ړo^B
    void Register(const string& name, const std::tr1::shared_ptr<void>& object) {
        instances.insert(name, object);
    }

    /// IuWFNg̎擾
    std::tr1::shared_ptr<void> Get(const string& name) {
        std::map<string, std::tr1::shared_ptr<void> >::iterator it = instances.find(name);
        if (it != instances.end()) {
            return (*it).second;
        }
        std::map<string, ComponentDef>::iterator it2 = componentDefs.find(name);
        if (it2 == componentDefs.end()) {
            return std::tr1::shared_ptr<void>();
        }
        std::tr1::shared_ptr<void> instance = Build(name, );
        instances.insert(name, instance);
        return instance;
    }

protected:
    /// IuWFNg̍\z
    std::tr1::shared_ptr<void> Build(const string& name, const ComponentDef& componentDef);
};

}
