#include "stdafx.h"
#include "DirObserver.h"
#include "Thread/Thread.h"
#include "Thread/ThreadPool.h"
#include "String/Encoding.h"
namespace Themis {

CDirObserver::CDirObserver() : m_hFolder(NULL) {
    threadValid = false;
}

CDirObserver::CDirObserver(const std::tr1::shared_ptr<IListener>& p): m_pListener(p), m_hFolder(NULL) {
    threadValid = false;
}

CDirObserver::~CDirObserver() {
    End();
}

void CDirObserver::Begin(const CFileName& oDirName, bool bSubFolder) {
    End();
    ResetChangeCount();
    m_oDirName = oDirName;
    m_hFolder = ::FindFirstChangeNotification(
        Encoding::GetDefault().GetBytes(oDirName.RemoveDirSeparator()).c_str(), bSubFolder?TRUE:FALSE,
        FILE_NOTIFY_CHANGE_FILE_NAME | FILE_NOTIFY_CHANGE_DIR_NAME |
        FILE_NOTIFY_CHANGE_ATTRIBUTES | FILE_NOTIFY_CHANGE_SIZE |
        FILE_NOTIFY_CHANGE_LAST_WRITE | FILE_NOTIFY_CHANGE_SECURITY
        );
    if (m_hFolder != NULL) {
        // XbhJn
        threadValid = true;
        thread = ThreadPool::CreateWorkerThread();
        thread->SetWork(std::tr1::bind(&CDirObserver::OnThread, this));
    }
}

void CDirObserver::End() {
    if (threadValid) {
        threadValid = false;
        if (m_hFolder != NULL) {
            ::FindCloseChangeNotification(m_hFolder);
            m_hFolder = NULL;
        }
        thread->Join();
    }
}

void CDirObserver::OnThread() {
    Thread::SetName(L"DirObserverThread");

    while (threadValid) {
        if (::WaitForSingleObject(m_hFolder, INFINITE) == WAIT_OBJECT_0) {
            // ύXۂꍇ
            m_nChange++;
            if (m_pListener != NULL) {
                m_pListener->OnChanged(*this);
            }
        }
        if (threadValid) {
            // ̕ύX҂
            BOOL b = ::FindNextChangeNotification(m_hFolder);
            Themis_DebugAssertApi(b != FALSE);
        } else {
            break; // ύX҂I
        }
    }
}

}
