#include "stdafx.h"
#include <math.h>
#include "Config.h"
#include "Graphics.h"
#include "ImageFilter.h"
namespace Themis {

CImageFilter::CImageFilter(CConfig& rConfig) : m_rConfig(rConfig), m_bFiltering(false) {
}

CImageFilter::~CImageFilter() {
}

bool CImageFilter::IsFiltering() const {
    return m_bFiltering;
}

void CImageFilter::Filter(std::tr1::shared_ptr<CPlaneEx>& pPlane, std::tr1::shared_ptr<CPlaneEx>& pTempPlane, bool& bHasAlpha) {
    bool bFilter = false;
    if (m_rConfig.Bool(L"Filter")) {
        if (m_rConfig.Bool(L"FilterAlpha") && pPlane->GetBpp() == 32) {
            if (!m_rConfig.Bool(L"FilterAlphaCheck") || bHasAlpha) {
                pTempPlane.swap(pPlane);
                pPlane->SetSize(pTempPlane->GetSize());
                pPlane->Fill(m_rConfig.Color(L"FilterAlphaBg"));
                pPlane->BltAlpha(static_cast<CPlane32&>(*pTempPlane), CPoint(), pTempPlane->GetSize().ToRect());
                bFilter = true;
                bHasAlpha = false; // Kp疳Ȃ̂ŁBB
            }
        }
        if (m_rConfig.Num(L"FilterGray") != 0) {
            if (m_rConfig.Num(L"FilterGray") == 1) {
                pPlane->GrayFill(pPlane->GetSize().ToRect());
            } else {
                Themis_DebugAssert(m_rConfig.Num(L"FilterGray") == 2);
                pPlane->GrayFillStrict(pPlane->GetSize().ToRect());
            }
            bFilter = true;
        }
        if (m_rConfig.Bool(L"FilterLevelRev")) {
            pPlane->LevelRevFill(pPlane->GetSize().ToRect(), m_rConfig.Num(L"FilterLevelRevLow"), m_rConfig.Num(L"FilterLevelRevHigh"));
            bFilter = true;
        }
        if (m_rConfig.Bool(L"FilterGammaRev")) {
            pPlane->GammaRevFill(pPlane->GetSize().ToRect(), pow(10.0, (double(m_rConfig.Num(L"FilterGammaRevValue")) / 100.0)));
            bFilter = true;
        }
        if (m_rConfig.Bool(L"FilterBC")) {
            pPlane->BrightContrastFill(pPlane->GetSize().ToRect(), m_rConfig.Num(L"FilterBCBright"), m_rConfig.Num(L"FilterBCContrast"));
            bFilter = true;
        }
        if (m_rConfig.Bool(L"FilterMedian")) {
            pTempPlane.swap(pPlane);
            pPlane->SetSize(pTempPlane->GetSize());
            pPlane->MedianFilterBlt(*pTempPlane, CPoint(), pTempPlane->GetSize().ToRect());
            bFilter = true;
        }
        if (m_rConfig.Num(L"FilterSharp") != 0) {
            pTempPlane->SetSize(pPlane->GetSize());
            switch (m_rConfig.Num(L"FilterSharp")) {
            case 1:
                pTempPlane.swap(pPlane);
                pPlane->BefogBlt(*pTempPlane, CPoint(), pTempPlane->GetSize().ToRect());
                break;
            case 2:
                pTempPlane.swap(pPlane);
                pPlane->SharpBlt(*pTempPlane, CPoint(), pTempPlane->GetSize().ToRect());
                break;
            case 3:
                pTempPlane->BefogBlt(*pPlane, CPoint(), pPlane->GetSize().ToRect());
                pPlane->BefogBlt(*pTempPlane, CPoint(), pTempPlane->GetSize().ToRect());
                break;
            case 4:
                pTempPlane.swap(pPlane);
                pPlane->HardSharpBlt(*pTempPlane, CPoint(), pTempPlane->GetSize().ToRect());
                break;
            default:
                Themis_DebugWarning(L"FilterSharp̒l");
            }
            bFilter = true;
        }
        if (m_rConfig.Bool(L"FilterGauss")) {
            pTempPlane.swap(pPlane);
            pPlane->SetSize(pTempPlane->GetSize());
            pPlane->GaussBefogBlt(*pTempPlane, CPoint(), pTempPlane->GetSize().ToRect(), m_rConfig.Num(L"FilterGaussValue") / 100.0);
            bFilter = true;
        }
        if (m_rConfig.Bool(L"FilterUnsharp")) {
            pTempPlane->SetSize(pPlane->GetSize());
            pTempPlane->GaussBefogBlt(*pPlane, CPoint(), pTempPlane->GetSize().ToRect(), m_rConfig.Num(L"FilterGaussValue") / 100.0);
            pPlane->UnsharpMaskBlt(*pPlane, *pTempPlane, CPoint(), pPlane->GetSize().ToRect(), CPoint()); 
            bFilter = true;
        }
        if (m_rConfig.Bool(L"FilterNeg")) {
            pPlane->ReverseFill(pPlane->GetSize().ToRect());
            bFilter = true;
        }
        if (m_rConfig.Num(L"FilterRot1") != 0) {
            pTempPlane.swap(pPlane);
            pPlane->SetSize(pTempPlane->GetSize());
            switch (m_rConfig.Num(L"FilterRot1")) {
            case 1:
                pPlane->MirrorHBlt(*pTempPlane, CPoint(), pTempPlane->GetSize().ToRect());
                break;
            case 2:
                pPlane->MirrorVBlt(*pTempPlane, CPoint(), pTempPlane->GetSize().ToRect());
                break;
            case 3:
                pPlane->Rot180Blt(*pTempPlane, CPoint(), pTempPlane->GetSize().ToRect());
                break;
            default:
                Themis_DebugWarning(L"FilterRot1̒l");
            }
            bFilter = true;
        }
        if (m_rConfig.Num(L"FilterRot2") != 0) {
            pTempPlane.swap(pPlane);
            pPlane->SetSize(CPlaneEx::GetRot90Size(pTempPlane->GetSize()));
            switch (m_rConfig.Num(L"FilterRot2")) {
            case 1:
                pPlane->Rot90ClockwiseBlt(*pTempPlane, CPoint(), pTempPlane->GetSize().ToRect());
                break;
            case 2:
                pPlane->Rot90AntiClockwiseBlt(*pTempPlane, CPoint(), pTempPlane->GetSize().ToRect());
                break;
            default:
                Themis_DebugWarning(L"FilterRot2̒l");
            }
            bFilter = true;
        }
    }
    m_bFiltering = bFilter;
}

}
