#pragma once
namespace Themis {
namespace Detail {

struct SPixelAverationData {
    boost::scoped_array<int> index;
    boost::scoped_array<int> tap;
    boost::scoped_array<int*> weight;
    boost::scoped_array<int> memory;
    void Set(int nMaxTap, int nCount) {
        index.reset(new int[nCount]);
        tap.reset(new int[nCount]);
        weight.reset(new int*[nCount]);
        memory.reset(new int[nMaxTap * nCount]);
        int* p = memory.get();
        for (int i = 0; i < nCount; i++) {
            weight[i] = p;
            p += nMaxTap;
        }
    }
};

inline void SetPixelAverationParamL(int nDstL, int nSrcL, SPixelAverationData& oData) {
    oData.Set(2, nDstL);
    const double dRate = double(nSrcL) / nDstL;
    for (int i = 0; i < nDstL; i++) {
        const double dPos1 = i * dRate;
        const int nPos1 = int(floor(dPos1));
        const double dWeight1 = dPos1 - floor(dPos1);
        const int nWeight1 = static_cast<int>(floor(dWeight1 * 0x10000 + 0.5));
        oData.index[i] = nPos1;
        if (nWeight1 == 0 || nPos1 == nSrcL - 1) {
            oData.tap[i] = 1;
            oData.weight[i][0] = 0x10000;
        } else {
            oData.tap[i] = 2;
            oData.weight[i][0] = 0x10000 - nWeight1;
            oData.weight[i][1] = nWeight1;
        }
    }
}

inline void SetPixelAverationParamD(int nDstL, int nSrcL, SPixelAverationData& oData) {
    const int nMaxTap = (nSrcL + nDstL - 1) / nDstL + 1;
    oData.Set(nMaxTap, nDstL);
    boost::scoped_array<double> work(new double[nMaxTap]);
    const double dRate = double(nSrcL) / nDstL;
    for (int i = 0; i < nDstL; i++) {
        const double dPos1 = i * dRate;
        const double dPos2 = dPos1 + dRate;
        const int nPos1 = int(floor(dPos1));
        const int nPos2 = int(ceil(dPos2)) - 1;
        const int nTap = nPos2 - nPos1 + 1;
        Themis_DebugAssert(0 < nTap);
        Themis_DebugAssert(nTap <= nMaxTap);
        double sum = 0.0;
        oData.index[i] = nPos1;
        oData.tap[i] = nTap;
        for (int j = 0; j < nTap; j++) {
            if (j == 0) {
                work[j] = 1.0 - (dPos1 - floor(dPos1));
            } else if (j == nTap - 1) {
                work[j] = dPos2 - nPos2;
            } else {
                work[j] = 1.0;
            }
            sum += work[j];
        }
        for (int j = 0; j < nTap; j++) {
            oData.weight[i][j] = int((work[j] / sum) * 0x10000);
        }
        if (oData.weight[i][nTap - 1] == 0) { // Ô(?)ŌオOȂQƂȂ
            oData.tap[i]--;
        }
    }
}

inline void SetPixelAverationParam(int nDstL, int nSrcL, SPixelAverationData& oData) {
    if (nSrcL <= nDstL) {
        SetPixelAverationParamL(nDstL, nSrcL, oData);
    } else {
        SetPixelAverationParamD(nDstL, nSrcL, oData);
    }
}

inline void PixelAverationH(uint32_t* pDst, int nDstPitch, uint32_t* pSrc, int nSrcPitch,
int nDstW, int nSrcW, int startY, int nHeight, const SPixelAverationData& oData) {
    int* tap = oData.tap.get();
    int* index = oData.index.get();
    int** weight = oData.weight.get();
    pDst += nDstPitch * startY;
    pSrc += nSrcPitch * startY;
    for (int y = startY; y < nHeight; y++) {
        for (int x = 0; x < nDstW; x++) {
            int dr = 0, dg = 0, db = 0;
            const int nTap = oData.tap[x];
            const int nPos = oData.index[x];
            for (int i = 0; i < nTap; i++) {
                const uint32_t dwSrc = pSrc[nPos + i];
                Themis_PlaneGetBGR32(dwSrc, r, g, b);
                dr += r * oData.weight[x][i];
                dg += g * oData.weight[x][i];
                db += b * oData.weight[x][i];
            }
            dr = (dr + 0x8000) >> 16;
            dg = (dg + 0x8000) >> 16;
            db = (db + 0x8000) >> 16;
            pDst[x] = Detail::BGR32(Detail::GetFixByte(dr), Detail::GetFixByte(dg), Detail::GetFixByte(db));
        }
        pDst += nDstPitch;
        pSrc += nSrcPitch;
    }
}
inline void PixelAverationHAlpha(uint32_t* pDst, int nDstPitch, uint32_t* pSrc, int nSrcPitch,
int nDstW, int nSrcW, int startY, int nHeight, const SPixelAverationData& oData) {
    int* tap = oData.tap.get();
    int* index = oData.index.get();
    int** weight = oData.weight.get();
    pDst += nDstPitch * startY;
    pSrc += nSrcPitch * startY;
    for (int y = startY; y < nHeight; y++) {
        for (int x = 0; x < nDstW; x++) {
            int da = 0, dr = 0, dg = 0, db = 0;
            const int nTap = oData.tap[x];
            const int nPos = oData.index[x];
            for (int i = 0; i < nTap; i++) {
                const uint32_t dwSrc = pSrc[nPos + i];
                Themis_PlaneGetBGRA32(dwSrc, a, r, g, b);
                da += a * oData.weight[x][i];
                dr += r * oData.weight[x][i];
                dg += g * oData.weight[x][i];
                db += b * oData.weight[x][i];
            }
            da = (da + 0x8000) >> 16;
            dr = (dr + 0x8000) >> 16;
            dg = (dg + 0x8000) >> 16;
            db = (db + 0x8000) >> 16;
            pDst[x] = Detail::BGRA32(Detail::GetFixByte(da), Detail::GetFixByte(dr), Detail::GetFixByte(dg), Detail::GetFixByte(db));
        }
        pDst += nDstPitch;
        pSrc += nSrcPitch;
    }
}

inline void PixelAverationV(uint32_t* pDst, int nDstPitch, uint32_t* pSrc, int nSrcPitch,
int nWidth, int startY, int endDstY, int nSrcH, const SPixelAverationData& oData) {
    int* tap = oData.tap.get();
    int* index = oData.index.get();
    int** weight = oData.weight.get();
    pDst += nDstPitch * startY;
    for (int y = startY; y < endDstY; y++) {
        for (int x = 0; x < nWidth; x++) {
            int dr = 0, dg = 0, db = 0;
            const int nTap = oData.tap[y];
            const int nPos = x + nSrcPitch * oData.index[y];
            for (int i = 0; i < nTap; i++) {
                const uint32_t dwSrc = pSrc[nPos + nSrcPitch * i];
                Themis_PlaneGetBGR32(dwSrc, r, g, b);
                dr += r * oData.weight[y][i];
                dg += g * oData.weight[y][i];
                db += b * oData.weight[y][i];
            }
            dr = (dr + 0x8000) >> 16;
            dg = (dg + 0x8000) >> 16;
            db = (db + 0x8000) >> 16;
            pDst[x] = Detail::BGR32(Detail::GetFixByte(dr), Detail::GetFixByte(dg), Detail::GetFixByte(db));
        }
        pDst += nDstPitch;
    }
}
inline void PixelAverationVAlpha(uint32_t* pDst, int nDstPitch, uint32_t* pSrc, int nSrcPitch,
int nWidth, int startY, int endDstY, int nSrcH, const SPixelAverationData& oData) {
    int* tap = oData.tap.get();
    int* index = oData.index.get();
    int** weight = oData.weight.get();
    pDst += nDstPitch * startY;
    for (int y = startY; y < endDstY; y++) {
        for (int x = 0; x < nWidth; x++) {
            int da = 0, dr = 0, dg = 0, db = 0;
            const int nTap = oData.tap[y];
            const int nPos = x + nSrcPitch * oData.index[y];
            for (int i = 0; i < nTap; i++) {
                const uint32_t dwSrc = pSrc[nPos + nSrcPitch * i];
                Themis_PlaneGetBGRA32(dwSrc, a, r, g, b);
                da += a * oData.weight[y][i];
                dr += r * oData.weight[y][i];
                dg += g * oData.weight[y][i];
                db += b * oData.weight[y][i];
            }
            da = (da + 0x8000) >> 16;
            dr = (dr + 0x8000) >> 16;
            dg = (dg + 0x8000) >> 16;
            db = (db + 0x8000) >> 16;
            pDst[x] = Detail::BGRA32(Detail::GetFixByte(da), Detail::GetFixByte(dr), Detail::GetFixByte(dg), Detail::GetFixByte(db));
        }
        pDst += nDstPitch;
    }
}

}//namespace Detail
}
