#include <math.h>       // sin, cos
#include "../../Windows/CpuId.h"
#include "../GDI/DC.h"
namespace Themis {

void CPlaneEx::BrightContrastFill(const CRect& rcDst, int nBright, int nContrast) {
    uint8_t abyTable[0x100];
    MakeBrightContrastTable(abyTable, nBright, nContrast);
    LevelRevTableFill(rcDst, abyTable);
}

void CPlaneEx::GammaRevFill(const CRect& rcDst, double dGamma) {
    uint8_t abyTable[0x100];
    MakeGammaRevTable(abyTable, dGamma);
    LevelRevTableFill(rcDst, abyTable);
}

void CPlaneEx::LevelRevFill(const CRect& rcDst, int nLow, int nHigh) {
    if (nLow <= nHigh) {
        uint8_t abyTable[0x100];
        MakeLevelRevTable(abyTable, nLow, nHigh);
        LevelRevTableFill(rcDst, abyTable);
    } else {
        uint8_t abyTable[0x100];
        MakeLevelRevTable(abyTable, nHigh, nLow);
        ReverseFill(rcDst);
        LevelRevFill(rcDst, nHigh, nLow);
    }
}

void CPlaneEx::MakeBrightContrastTable(uint8_t abyTable[0x100], int nBright, int nContrast) {
    Themis_DebugAssert(-256 <= nBright);
    Themis_DebugAssert(nBright <= 256);
    Themis_DebugAssert(-256 <= nContrast);
    Themis_DebugAssert(nContrast <= 256);
    if (nContrast != 0) {
      #if 0
        COLORREF rgbAve = GetAverageColor(rcDst);
        const int nOrg = (GetRValue(rgbAve) + GetGValue(rgbAve) + GetBValue(rgbAve)) / 3;
      #else
        const int nOrg = 127;
      #endif
        for (int i = 0; i < 0x100; i++) {
            int t = i + (i - nOrg) * nContrast / 256 + nBright;
            abyTable[i] = Detail::GetFixByte(t);
        }
    } else { // RgXg␳
        for (int i = 0; i < 0x100; i++) {
            abyTable[i] = Detail::GetFixByte(i + nBright);
        }
    }
}

void CPlaneEx::MakeGammaRevTable(uint8_t abyTable[0x100], double dGamma) {
    Themis_DebugAssert(0.0 < dGamma);
    for (int i = 0; i < 0x100; i++) {
        double d = pow(double(i) / 256.0, 1.0 / dGamma) * 256.0;
        Themis_PlaneCheckByte(int(d));
        abyTable[i] = uint8_t(d);
    }
}

void CPlaneEx::MakeLevelRevTable(uint8_t abyTable[0x100], int nLow, int nHigh) {
    Themis_DebugAssert(nLow <= nHigh);
    Themis_DebugAssert(0 <= nLow);
    Themis_DebugAssert(nHigh <= 0xff);
    std::fill(abyTable, abyTable + nLow, 0);
    for (int i = nLow; i < nHigh; i++) {
        abyTable[i] = static_cast<uint8_t>(0xff * (i - nLow) / (nHigh - nLow));
    }
    std::fill(abyTable + nHigh, abyTable + 0x100, 0xff);
}

//////////////////////////////////////////////////////////////////////

void CPlane8::ReverseFill(CRect rcDst) {
    if (CClipper(GetSize()).Clipping(rcDst)) {
        Themis_PlaneLock1(*this, rcDst.GetPoint(), pDst, nDstPitch);
        Themis_PlaneGetSize(rcDst.GetSize(), nDstW, nDstH);
        for (int y = 0; y < nDstH; y++) {
            for (int x = 0; x < nDstW; x++) {
                pDst[x] ^= 0xff;
            }
            pDst += nDstPitch;
        }
    }
}

void CPlane8::GetHistogram(int anHistogram[0x100], CRect rcDst) {
    std::fill(anHistogram, anHistogram + 0x100, 0);
    if (CClipper(GetSize()).Clipping(rcDst)) {
        Themis_PlaneLock1(*this, rcDst.GetPoint(), pDst, nDstPitch);
        Themis_PlaneGetSize(rcDst.GetSize(), nDstW, nDstH);
        for (int y = 0; y < nDstH; y++) {
            for (int x = 0; x < nDstW; x++) {
                anHistogram[pDst[x]]++;
            }
            pDst += nDstPitch;
        }
    }
}

}
