#include <math.h>
namespace Themis {

bool CClipper::Clipping(CRect& rcDst) const {
    rcDst.SetIntersect(GetRect());
    return rcDst.IsValid();
}

/// `{]p(HDCpv[TCYw)
bool CClipper::Clipping(CPoint& ptDst, CRect& rcSrc) const {
    if (!rcSrc.IsValid()) {
        return false;
    }
    //]␳
    CRect rcDst(ptDst, rcSrc.GetSize());
    if (IsValid()) {
        rcDst.SetIntersect(GetRect());
    }
    rcSrc = CRect((rcSrc.GetPoint() + rcDst.GetPoint() - ptDst), rcDst.GetSize());
    ptDst = rcDst.GetPoint();
    return rcSrc.IsValid();
}

/// `Lk]p(HDCpv[TCYw)
bool CClipper::Clipping(CRect& rcDst, CRect& rcSrc) const {
    if (!rcDst.IsValid() || !rcSrc.IsValid()) {
        return false;
    }
    double fRateX = static_cast<double>(rcDst.GetWidth()) / rcSrc.GetWidth();
    double fRateY = static_cast<double>(rcDst.GetHeight()) / rcSrc.GetHeight();
    //]␳
    CRect rcDstOld = rcDst;
    if (IsValid()) {
        rcDst.SetIntersect(GetRect());
    }
    CSize szDstGap1 = rcDstOld.GetPoint() - rcDst.GetPoint();   //Ə
    CSize szDstGap2 = CSize(rcDstOld.right - rcDst.right, rcDstOld.bottom - rcDst.bottom);  //EƉ
    rcSrc.left -= static_cast<int>(floor(szDstGap1.cx / fRateX + 0.5));
    rcSrc.top -= static_cast<int>(floor(szDstGap1.cy / fRateY + 0.5));
    rcSrc.right -= static_cast<int>(floor(szDstGap2.cx / fRateX + 0.5));
    rcSrc.bottom -= static_cast<int>(floor(szDstGap2.cy / fRateY + 0.5));
    return rcDst.IsValid() && rcSrc.IsValid();
}

bool CClipper::Clipping(const CSize& szDst, const CSize& szSrc, CPoint& ptDst, CRect& rcSrc) const {
    if (!rcSrc.IsValid()) {
        return false;
    }
    //]␳
    CRect rcDst(ptDst, rcSrc.GetSize());
    rcDst.SetIntersect(CRect(szDst));
    if (IsValid()) {
        rcDst.SetIntersect(GetRect());
    }
    rcSrc = CRect((rcSrc.GetPoint() + rcDst.GetPoint() - ptDst), rcDst.GetSize());
    ptDst = rcDst.GetPoint();
    //]␳
    CPoint ptSrcOld = rcSrc.GetPoint();
    rcSrc.SetIntersect(CRect(szSrc));
    ptDst = ptDst - ptSrcOld + rcSrc.GetPoint();
    return rcSrc.IsValid();
}

bool CClipper::Clipping(const CSize& szDst, const CSize& szSrc, CRect& rcDst, CRect& rcSrc) const {
    if (!rcDst.IsValid() || !rcSrc.IsValid()) {
        return false;
    }
    double fRateX = static_cast<double>(rcDst.GetWidth()) / rcSrc.GetWidth();
    double fRateY = static_cast<double>(rcDst.GetHeight()) / rcSrc.GetHeight();
    //]␳
    CRect rcDstOld = rcDst;
    rcDst.SetIntersect(CRect(szDst));
    if (IsValid()) {
        rcDst.SetIntersect(GetRect());
    }
    CSize szDstGap1 = rcDstOld.GetPoint() - rcDst.GetPoint();   //Ə
    CSize szDstGap2 = CSize(rcDstOld.right - rcDst.right, rcDstOld.bottom - rcDst.bottom);  //EƉ
    rcSrc.left -= static_cast<int>(floor(szDstGap1.cx / fRateX + 0.5));
    rcSrc.top -= static_cast<int>(floor(szDstGap1.cy / fRateY + 0.5));
    rcSrc.right -= static_cast<int>(floor(szDstGap2.cx / fRateX + 0.5));
    rcSrc.bottom -= static_cast<int>(floor(szDstGap2.cy / fRateY + 0.5));
    //]␳
    CRect rcSrcOld = rcSrc;
    rcSrc.SetIntersect(CRect(szSrc));    //͂ݏo␳
    CSize szSrcGap1 = rcSrcOld.GetPoint() - rcSrc.GetPoint();   //Ə
    CSize szSrcGap2 = CSize(rcSrcOld.right - rcSrc.right, rcSrcOld.bottom - rcSrc.bottom);  //EƉ
    rcDst.left -= static_cast<int>(floor(szSrcGap1.cx * fRateX + 0.5));
    rcDst.top -= static_cast<int>(floor(szSrcGap1.cy * fRateY + 0.5));
    rcDst.right -= static_cast<int>(floor(szSrcGap2.cx * fRateX + 0.5));
    rcDst.bottom -= static_cast<int>(floor(szSrcGap2.cy * fRateY + 0.5));
    return rcDst.IsValid() && rcSrc.IsValid();
}

bool CClipper::ClippingRot90(const CSize& szDst, const CSize& szSrc, CPoint& ptDst, CRect& rcSrc) const {
    if (!rcSrc.IsValid()) {
        return false;
    }
    //]␳
    CRect rcDst(ptDst, rcSrc.GetHeight(), rcSrc.GetWidth());    //Dstw, hSrch, w
    rcDst.SetIntersect(CRect(szDst));
    if (IsValid()) {
        rcDst.SetIntersect(GetRect());
    }
    rcSrc = CRect((rcSrc.GetPoint() + rcDst.GetPoint() - ptDst), CSize(rcDst.GetHeight(), rcDst.GetWidth()));   //łcւ
    ptDst = rcDst.GetPoint();
    //]␳
    CPoint ptSrcOld = rcSrc.GetPoint();
    rcSrc.SetIntersect(CRect(szSrc));
    ptDst = ptDst - ptSrcOld + rcSrc.GetPoint();
    return rcDst.IsValid() && rcSrc.IsValid();
}

}
