namespace Themis {

namespace {
    static bool useTopDownDIB = true;
}
/// TopDownDIBĝ
bool CDib::IsUseTopDownDIB() { return useTopDownDIB; }
/// TopDownDIBĝ
void CDib::SetUseTopDownDIB(bool b) { useTopDownDIB = b; }

CDibImpl::CDibImpl(): m_hdc(NULL), m_hbmp(NULL), m_hbmpOld(NULL), m_pBmpInfo(NULL), m_pBits(NULL) {
}

CDibImpl::~CDibImpl() {
    Delete();
}

bool CDibImpl::IsCreated() const {
    return (m_hdc!=NULL && m_pBmpInfo!=NULL && m_hbmp!=NULL);
}

void CDibImpl::Create(const CSize& sz, int nBpp) {
    try {
        if (nBpp != 16 && nBpp != 24 && nBpp != 32) {
            Themis_DebugWarning(L"DIBBPPs");
        }
        if (IsCreated() && GetSize() == sz) {
            return;
        }
        Delete();

        m_pBmpInfo = CreateBMPINFO();
        if (m_pBmpInfo == NULL) { throw RuntimeException(L"摜ێ(DIB)̍쐬Ɏs܂"); }
        FillBmiHeader(m_pBmpInfo->bmiHeader, CSize(((sz.cx + 3) & ~3), sz.cy), nBpp); //͂S̔{
    //*
        m_hdc = ::CreateCompatibleDC(NULL);
        if (m_hdc==NULL) { throw Win32Exception(L"摜ێ(DIB)̍쐬Ɏs܂"); }
        m_hbmp = ::CreateDIBSection(m_hdc, m_pBmpInfo, DIB_RGB_COLORS, &m_pBits, NULL, 0);
        if (m_hbmp==NULL) { throw Win32Exception(L"摜ێ(DIB)̍쐬Ɏs܂"); }
    /*/
        HDC hdcTmp = ::GetDC(NULL);
        m_hbmp = ::CreateDIBSection(hdcTmp, m_pBmpInfo, DIB_RGB_COLORS, &m_pBits, NULL, 0);
        if (m_hbmp==NULL) { throw Win32Exception(L"摜ێ(DIB)̍쐬Ɏs܂"); }
        m_hdc = ::CreateCompatibleDC(hdcTmp);
        if (m_hdc==NULL) { throw Win32Exception(L"摜ێ(DIB)̍쐬Ɏs܂"); }
    //*/
        m_hbmpOld = reinterpret_cast<HBITMAP>(::SelectObject(m_hdc, m_hbmp));
        m_nWidth = sz.cx;
    } catch (std::exception&) {
        throw;
    }
}
void CDibImpl::Delete() {
    if (m_hdc != NULL) {
        if (m_hbmp != NULL) {
            ::SelectObject(m_hdc, m_hbmpOld);
            ::DeleteObject(m_hbmp);
            DeleteBMPINFO(m_pBmpInfo);
        }
        ::DeleteDC(m_hdc);
        m_hbmp = NULL;
        m_hdc = NULL;
    } else if (m_hbmp != NULL) {
        //邿Ƃ͍쐬ŝˁc
        ::DeleteObject(m_hbmp);
        m_hbmp = NULL;
        DeleteBMPINFO(m_pBmpInfo);
    }
}
BITMAPINFO* CDibImpl::CreateBMPINFO() {
    return static_cast<BITMAPINFO*>(operator new(sizeof(BITMAPINFOHEADER)));
}
void CDibImpl::DeleteBMPINFO(BITMAPINFO* pBmpInfo) {
    operator delete(pBmpInfo);
}
void CDibImpl::FillBmiHeader(BITMAPINFOHEADER& r, const CSize& sz, int nBpp) {
    r.biSize = sizeof(BITMAPINFOHEADER);
    r.biWidth = sz.cx;
    r.biHeight = CDib::IsUseTopDownDIB() ?
        -sz.cy :       //ォ牺Aʂۂz
        sz.cy;         //̓VntȔz
    r.biPlanes = 1;
    r.biBitCount = static_cast<WORD>(nBpp);
    r.biCompression = BI_RGB;
    r.biSizeImage = 0;          //w*h*bpp/bitof(uint8_t) H
    r.biXPelsPerMeter = 0;      //378096dpic
    r.biYPelsPerMeter = 0;
    r.biClrUsed = 0;
    r.biClrImportant = 0;
}

CSize CDibImpl::GetSize() const {
    if (IsCreated()) {
        return CSize(m_nWidth, abs(m_pBmpInfo->bmiHeader.biHeight));
    } else {
        return CSize(0, 0);
    }
}

pair<void*, int> CDibImpl::Lock() {
    ::GdiFlush();   //GDI`ɔf

    int nWidth = m_pBmpInfo->bmiHeader.biWidth;
    int nPitch = nWidth * (m_pBmpInfo->bmiHeader.biBitCount / 8);
    if (m_pBmpInfo->bmiHeader.biHeight <= 0) {
        return make_pair(m_pBits, nPitch);
    } else {
        void* pBits = static_cast<uint8_t*>(m_pBits) + nPitch * (m_pBmpInfo->bmiHeader.biHeight - 1);
        return make_pair(pBits, - nPitch);
    }
}

HDC CDibImpl::GdiLock() {
    return m_hdc;
}

}
