#include "stdafx.h"
#include "Config2.h"
#include "HotKey.h"
namespace Themis {

HotKeySetting::HotKeySetting() {
}

HotKeySetting::~HotKeySetting() {
}

void HotKeySetting::Clear() {
    boost::recursive_mutex::scoped_lock lock(mutex);
    m_oIdList.clear();
    m_oKeyList.clear();
}

void HotKeySetting::Remove(CKeyTable::EKey eKey, int nModifiers) {
    boost::recursive_mutex::scoped_lock lock(mutex);
    SKey oKey(eKey, nModifiers);
    std::map<SKey, int>::iterator it = m_oIdList.find(oKey);
    if (it != m_oIdList.end()) {
        m_oKeyList[(*it).second].erase(oKey);
        m_oIdList.erase(oKey);
    }
}

void HotKeySetting::Add(int nHotKeyId, CKeyTable::EKey eKey, int nModifiers) {
    boost::recursive_mutex::scoped_lock lock(mutex);
    SKey skey(eKey, nModifiers);
    std::map<SKey, int>::iterator it = m_oIdList.find(skey);
    if (it == m_oIdList.end()) {
        // ǉ
        m_oIdList.insert(make_pair(skey, nHotKeyId));
        m_oKeyList[nHotKeyId].insert(skey);
    } else {
        // ̂폜ǉ
        m_oIdList[skey] = nHotKeyId;
        m_oKeyList[(*it).second].erase(skey);
        m_oKeyList[nHotKeyId].insert(skey);
    }
}

void HotKeySetting::Add(int nHotKeyId, const string& str) {
    if (!string::IsNullOrEmpty(str)) {
        InnerAdd(nHotKeyId, str);
    }
}
void HotKeySetting::InnerAdd(int nHotKeyId, const string& str) {
    List<string> vct = str.Split(Array<string>(L", "));
    for (size_t i = 0, n = vct.size(); i < n; i++) {
        pair<CKeyTable::EKey, int> p = CKeyTable::GetKey(vct[i]);
        if (p.first != CKeyTable::EKey(-1)) {
            Add(nHotKeyId, p.first, p.second);
        }
    }
}

int HotKeySetting::GetId(CKeyTable::EKey eKey, int nModifiers) const {
    boost::recursive_mutex::scoped_lock lock(mutex);
    std::map<SKey, int>::const_iterator it = m_oIdList.find(SKey(eKey, nModifiers));
    if (it != m_oIdList.end()) {
        return (*it).second;
    }
    return -1;
}

string HotKeySetting::GetKeyStr(int nHotKeyId) const {
    string str;
    std::map<int, std::set<SKey> >::const_iterator p = m_oKeyList.find(nHotKeyId);
    if (p == m_oKeyList.end()) {
        return str;
    }
    for (std::set<SKey>::const_iterator it = (*p).second.begin(), end = (*p).second.end(); it != end; ++it) {
        if (string::IsNullOrEmpty(str)) {
            str = CKeyTable::GetKeyName((*it).eKey, (*it).nModifiers);
        } else {
            str += L", ";
            str += CKeyTable::GetKeyName((*it).eKey, (*it).nModifiers);
        }
    }
    return str;
}

void HotKeySetting::SaveToConfig(const CFileName& oFileName, HotKeySetting& defaultSetting) {
    CConfig2 config;
    config.Num(L"[ver]") << 2;
    // ftHgƈقȂĂ̂̂ݏ
    for (std::map<SKey, int>::iterator it = m_oIdList.begin(); it != m_oIdList.end(); ++it) {
        SKey key = (*it).first;
        int command = (*it).second;
        int defaultCommand = defaultSetting.GetId(key.eKey, key.nModifiers);
        if (command != defaultCommand) {
            config.Num(CKeyTable::GetKeyName(key.eKey, key.nModifiers)) << command;
        }
    }
    // ftHgɂ݂Ȃ̂(xԂȂ̂BB)
    for (std::map<SKey, int>::iterator it = defaultSetting.m_oIdList.begin(); it != defaultSetting.m_oIdList.end(); ++it) {
        SKey key = (*it).first;
        int defaultCommand = (*it).second;
        int command = GetId(key.eKey, key.nModifiers);
        if (command < 0) {
            config.Num(CKeyTable::GetKeyName(key.eKey, key.nModifiers)) << command;
        }
    }
    config.GetStorage().Save(oFileName);
}

namespace {
    bool InnerLoad(HotKeySetting& setting, CConfig2& config) {
     // setting.Clear();
        vector<pair<string, string> > vct = config.GetStorage().GetValues();
        for (size_t i = 0, n = vct.size(); i < n; i++) {
            setting.Add(AtoI(vct[i].first), vct[i].second);
        }
        return true;
    }
}
void HotKeySetting::LoadFromConfig(const CFileName& oFileName, HotKeySetting& defaultSetting) {
    // Ƃ肠ftHgݒ
    m_oIdList = defaultSetting.m_oIdList;
    m_oKeyList = defaultSetting.m_oKeyList;
    // ǂݍ
    CConfig2 config;
    if (config.GetStorage().Load(oFileName)) {
        if (config.GetStorage().Exists(L"[ver]")) {
            if (config.Num(L"[ver]").Get() == 2) {
                // ݒt@Ĉ㏑B
                vector<pair<string, string> > vct = config.GetStorage().GetValues();
                for (size_t i = 0, n = vct.size(); i < n; i++) {
                    if (vct[i].first == L"[ver]") continue;
                    int command = AtoI(vct[i].second);
                    if (0 <= command) {
                        Add(command, vct[i].first);
                    }
                }
                return;
            }
        } else {
            // o[W
            if (InnerLoad(*this, config)) {
                return;
            }
        }
    }
    // s̓ftHgēxB
    m_oIdList = defaultSetting.m_oIdList;
    m_oKeyList = defaultSetting.m_oKeyList;
}

CHotKey::CHotKey(const std::tr1::shared_ptr<IOwner>& pOwner): m_pOwner(pOwner) {
}

CHotKey::~CHotKey() {
}

bool CHotKey::HookMsg(CMsgHookPrms& rPrms) {
    switch (rPrms.GetMsg()) {
    case WM_SYSKEYDOWN:
    case WM_KEYDOWN: {
        boost::recursive_mutex::scoped_lock lock(mutex);
        CKeyTable::EKey e = CKeyTable::VKeyToEKey(static_cast<UINT>(rPrms.GetWPrm()));
        if (e == CKeyTable::EKey(-1)) {
            return false;
        }
        return OnInput(e, CKeyTable::GetModifiers());
    }
    case WM_LBUTTONUP:
        if (rPrms.GetWPrm() & MK_RBUTTON) {
            return OnInput(CKeyTable::eKeyRLClick, CKeyTable::GetModifiers());
        } else {
            return OnInput(CKeyTable::eKeyLClick, CKeyTable::GetModifiers());
        }
    case WM_RBUTTONUP:
        if (rPrms.GetWPrm() & MK_LBUTTON) {
            return OnInput(CKeyTable::eKeyLRClick, CKeyTable::GetModifiers());
        } else {
            return OnInput(CKeyTable::eKeyRClick, CKeyTable::GetModifiers());
        }
    case WM_MBUTTONUP:
        return OnInput(CKeyTable::eKeyMClick, CKeyTable::GetModifiers());
    case 0x020c://WM_XBUTTONUP
        if (rPrms.GetWPrmHiWord() & 0x0001) { //XBUTTON1
            return OnInput(CKeyTable::eKey4Click, CKeyTable::GetModifiers());
        } else if (rPrms.GetWPrmHiWord() & 0x0002) { //XBUTTON2
            return OnInput(CKeyTable::eKey5Click, CKeyTable::GetModifiers());
        }
    case WM_LBUTTONDBLCLK:
        return OnInput(CKeyTable::eKeyLDouble, CKeyTable::GetModifiers());
    case WM_RBUTTONDBLCLK:
        return OnInput(CKeyTable::eKeyRDouble, CKeyTable::GetModifiers());
    case WM_MBUTTONDBLCLK:
        return OnInput(CKeyTable::eKeyMDouble, CKeyTable::GetModifiers());
    case 0x020d://WM_XBUTTONDBLCLK
        if (rPrms.GetWPrmHiWord() & 0x0001) { //XBUTTON1
            return OnInput(CKeyTable::eKey4Double, CKeyTable::GetModifiers());
        } else if (rPrms.GetWPrmHiWord() & 0x0002) { //XBUTTON2
            return OnInput(CKeyTable::eKey5Double, CKeyTable::GetModifiers());
        }
    case WM_MOUSEWHEEL: {
        int nDelta = GET_WHEEL_DELTA_WPARAM(rPrms.GetWPrm());
        // WHEEL_DELTAŊZĉ񐔕ĂяoׂcH
        if (nDelta <= -WHEEL_DELTA) {
            return OnInput(CKeyTable::eKeyMDown, CKeyTable::GetModifiers());
        } else if (WHEEL_DELTA <= nDelta) {
            return OnInput(CKeyTable::eKeyMUp, CKeyTable::GetModifiers());
        } else {
            return false;
        }
    }
    default:
        return false;
    }
}

bool CHotKey::OnInput(CKeyTable::EKey eKey, int nModifier) {
    boost::recursive_mutex::scoped_lock lock(mutex);
    int nHotKeyId = GetId(eKey, nModifier);
    if (nHotKeyId == -1) {
        return false;
    }
    m_pOwner->OnHotKey(this, nHotKeyId);
    return true;
}

}
