#include "stdafx.h"
#include <new>
#include "ReuseAllocator.h"
namespace Themis {

CReuseAllocator::CReuseAllocator() {
  #if Themis_ReuseAllocatorLog
    m_nLeakCount = 0;
    m_nMAllocCount = 0;
    m_nReuseCount = 0;
  #endif
}

CReuseAllocator::~CReuseAllocator() {
    Garbage();
  #if Themis_ReuseAllocatorLog
    Themis_DebugAlwaysLog(L"------ CReuseAllocator ------");
    Themis_DebugAlwaysLog(L"[N^fF" + ItoA(m_nLeakCount));
    Themis_DebugAlwaysLog(L"mیF" + ItoA(m_nMAllocCount));
    Themis_DebugAlwaysLog(L"ėp񐔁@@F" + ItoA(m_nReuseCount));
    Themis_DebugAlwaysLog(L"-----------------------------");
  #endif
}

namespace {
    inline size_t GetMemorySize(void* p) {
        return ((unsigned*)p)[-1];
    }
}

void* CReuseAllocator::Alloc(size_t nMin) {
    boost::mutex::scoped_lock lock(mutex);
    for (size_t i = 0, n = m_vctList.size(); i < n; i++) {
        size_t memorySize = GetMemorySize(m_vctList[i]);
        if (nMin <= memorySize && memorySize < nMin * 2) {
            // ܂fĴ͍ėp߂ƂC
          #if Themis_ReuseAllocatorLog
            m_nReuseCount++;
          #endif
            void* p = m_vctList[i];
            m_vctList.erase(m_vctList.begin() + i);
            return p;
        }
    }
    size_t nMemorySize = nMin + sizeof(size_t);
    nMemorySize = (nMemorySize + 7) & ~7; // Kɒ[
    void* pMemory = malloc(nMemorySize);
    if (pMemory == NULL) {
        throw std::bad_alloc();
    }
  #if Themis_ReuseAllocatorLog
    m_nLeakCount++;
    m_nMAllocCount++;
  #endif
    ((unsigned*)pMemory)[0] = static_cast<unsigned>(nMemorySize - sizeof(size_t));
    return &((unsigned*)pMemory)[1];
}

void CReuseAllocator::Free(void* p) {
    boost::mutex::scoped_lock lock(mutex);
    size_t nSize = GetMemorySize(p);
    size_t i = 0;
    for (size_t n = m_vctList.size(); i < n; i++) {
        if (nSize < GetMemorySize(m_vctList[i])) {
            break;
        }
    }
    m_vctList.insert(m_vctList.begin() + i, p);
}

size_t CReuseAllocator::GetCapacity(void* p) {
    return GetMemorySize(p);
}

void CReuseAllocator::Garbage() {
    vector<void*> vctCopy;
    {
        boost::mutex::scoped_lock lock(mutex);
        vctCopy.swap(m_vctList);
    }
    for (size_t i = 0, n = vctCopy.size(); i < n; i++) {
      #if Themis_ReuseAllocatorLog
        m_nLeakCount--;
      #endif
        free(&((unsigned*)vctCopy[i])[-1]);
    }
}

}
