#include "stdafx.h"
#include "Windows/System.h"
#include "Display.h"
#include "Window.h"
#include "LocalExclusion.h"
#include "ScreenChanger.h"
namespace Themis {

CScreenChanger::CScreenChanger(CWindow& rWindow):
m_rWindow(rWindow), m_nBpp(-1), m_bShowMenu(false), m_eState(eStateWindowMode),    //ŏ̓EBhE[h
m_bHideCursor(false), m_bDeleted(false), minOnDeactive(true) {
    m_pExclusion.reset(new CLocalExclusion);
    m_pMenubar.reset(new CWindow);
    m_pMenubar->GetHook().Add(this);
    CWindowParameter oParameter;
    oParameter.SetActiveOnCreate(false);
    oParameter.GetStyle() = WS_POPUP;
    oParameter.GetExStyle() = WS_EX_TOOLWINDOW;
    m_pMenubar->CreateWithThread(oParameter);
    m_pMenubar->SetSize(CSize(0, ::GetSystemMetrics(SM_CYMENU)));
    m_pMenubar->SetPos(CPoint(0, - ::GetSystemMetrics(SM_CYMENU)));
    m_rWindow.GetHook().Add(this);
    SetSizeWindow();
}

CScreenChanger::~CScreenChanger() {
    Delete();
}

void CScreenChanger::SetHideCursor(bool b) {
    m_bHideCursor = b;
    if (m_eState == eStateFullScreen) {
        if (b) {
            m_rWindow.HideCursor();
        } else {
            m_rWindow.ShowCursor();
        }
    }
}

void CScreenChanger::SetSizeWindow() {
    m_szScreen = m_rWindow.GetClientSize();
}

void CScreenChanger::SetSizeScreen() {
 // m_szScreen.cx = ::GetSystemMetrix(SM_CXSCREEN);
 // m_szScreen.cy = ::GetSystemMetrix(SM_CYSCREEN);
    m_szScreen = CDisplay::GetCurrentMode().oSize;
}

HMENU CScreenChanger::GetMenu() const {
    return m_eState == eStateFullScreen ? m_pMenubar->GetMenu() : m_rWindow.GetMenu();
}

void CScreenChanger::SetMenu(HMENU hMenu) {
    HideMenubar(); //蔲CoOh~[u
    return m_eState == eStateFullScreen ? m_pMenubar->SetMenu(hMenu) : m_rWindow.SetMenu(hMenu);
}

void CScreenChanger::SetMenu(int nMenu) {
    HideMenubar(); //蔲CoOh~[u
    return m_eState == eStateFullScreen ? m_pMenubar->SetMenu(nMenu) : m_rWindow.SetMenu(nMenu);
}

void CScreenChanger::DeleteMenu() {
    HideMenubar(); //蔲CoOh~[u
    return m_eState == eStateFullScreen ? m_pMenubar->DeleteMenu() : m_rWindow.DeleteMenu();
}

void CScreenChanger::DrawMenubar() {
    ::DrawMenuBar(m_eState == eStateFullScreen ? m_pMenubar->GetHandle() : m_rWindow.GetHandle());
}

/// Window[hɍőOʕ\̂ǂ
void CScreenChanger::SetTopMost(bool b) {
    if (IsFullScreen()) {
        if (b) {
            m_dwExStyleOld |= WS_EX_TOPMOST;
        } else {
            m_dwExStyleOld &= ~WS_EX_TOPMOST;
        }
    } else {
        m_rWindow.SetTopMost(b);
    }
}
/// Window[hɍőOʕ\Ȃtrue
bool CScreenChanger::IsTopMost() const {
    if (IsFullScreen()) {
        return (m_dwExStyleOld & WS_EX_TOPMOST) != 0;
    } else {
        return m_rWindow.IsTopMost();
    }
}

void CScreenChanger::ChangeMode() {
    IsFullScreen() ? ToWindowMode() : ToFullScreen();
}

void CScreenChanger::ToFullScreen() {
    CLocalExclusion::CLock oLock(*m_pExclusion);
    if (oLock.IsLocked()) {
        if (m_eState != eStateFullScreen) {
            m_eState = eStateFullScreen; // WM_SIZEƂoOȂ悤ɁAtO͐ɕύXĂ܂ĂB
            //ʐ؂ւJ[\B
            try {
                ChangeWindowState();
                ChangeDisplay();
            } catch (std::exception&) {
                RestoreWindowState();
                m_eState = eStateWindowMode;
                throw;
            }
            if (m_bHideCursor) {
                m_rWindow.HideCursor();
            }
            HideMenubar(); // j[o[͈ꉞBĂ
            m_rWindow.Update();
        }
    }
}

void CScreenChanger::ChangeWindowState() {
    //EBhEʒuEX^C̕ۑ
    m_nStateOld =
        m_rWindow.IsMaximized() ? 1 :
        m_rWindow.IsMinimized() ? 2 : 0;
    if (m_nStateOld != 0) {
        m_rWindow.ShowRestore();
    }
    m_rcOld = m_rWindow.GetRect();
    m_dwStyleOld = m_rWindow.GetStyle();
    m_dwExStyleOld = m_rWindow.GetExStyle();
 // m_hwndParentOld = m_rWindow.GetParent();

    m_rWindow.SetForeground();
  #ifdef NDEBUG
    m_rWindow.SetTopMost(true);
  #endif

    //烁j[o[OAj[o[ɃZbg
    HMENU hMenu = m_rWindow.GetMenu();
    if (hMenu != NULL) {
        m_rWindow.UnsetMenu();
        m_pMenubar->SetMenu(hMenu);
        m_pMenubar->Hide();
    } else {
        m_pMenubar->SetRect(CRect(-1, -1, 0, 0));
    }

    //CʑŜ֍L
    m_rWindow.SetStyle(WS_POPUP | WS_VISIBLE);
  #ifdef NDEBUG
    m_rWindow.SetExStyle(WS_EX_TOPMOST);
  #else
    m_rWindow.SetExStyle(0);
  #endif
 /* HWND hwndTray = ::FindWindow(L"Shell_TrayWnd", NULL);
    if (hwndTray != NULL && ::FindWindowEx(hwndTray, NULL, L"TrayNotifyWnd", NULL) != NULL) {
        // ꉞ딚΍ȒPɂƂB
        m_rWindow.SetParent(hwndTray);
    } */
    m_rWindow.SetPos(CPoint());
    m_rWindow.SetSize(m_szScreen);
}

void CScreenChanger::RestoreWindowState() {
    //j[烁j[o[O
    HMENU hMenu = NULL;
    if (m_pMenubar->Exists()) {
        hMenu = m_pMenubar->GetMenu();
        if (hMenu != NULL) {
            m_pMenubar->UnsetMenu();
        }
    }
    // C𑋃[hpɖ߂
    if (m_rWindow.Exists()) {
        if (hMenu != NULL) {
            m_rWindow.SetMenu(hMenu);
        }
     // m_rWindow.SetParent(m_hwndParentOld);
        m_rWindow.SetExStyle(m_dwExStyleOld);
        m_rWindow.SetStyle(m_dwStyleOld);
        m_rWindow.SetTopMost((m_dwExStyleOld & WS_EX_TOPMOST) != 0);
        m_rWindow.SetRect(m_rcOld);
        switch (m_nStateOld) {
        default:
        case 0: m_rWindow.ShowRestore(); break;
        case 1: m_rWindow.Maximize(); break;
        case 2: m_rWindow.Minimize(); break;
        }
        m_rWindow.Update();
    }
}

void CScreenChanger::ChangeDisplay() {
    CDisplay::SVideo v;
    v.oSize = m_szScreen;
    if (m_nBpp < 0) {
        v.nBpp = CDisplay::SVideo::nDefault;
    } else {
        v.nBpp = m_nBpp;
        if (!CDisplay::IsSupportedMode(v)) {
            v.nBpp = CDisplay::SVideo::nDefault;    //}uƂāBB
        }
    }
    CDisplay::SetTmp(v);
}

void CScreenChanger::ToWindowMode() {
    CLocalExclusion::CLock oLock(*m_pExclusion);
    if (oLock.IsLocked()) {
        if (m_eState != eStateWindowMode) {
            HideMenubar(); // Ô߁B
            if (m_eState != eStateTempWindowMode) {
                if (CDisplay::Restore()) {
                    ::InvalidateRect(NULL, NULL, FALSE);
                }
                if (m_bHideCursor) {
                    m_rWindow.ShowCursor();
                }
                RestoreWindowState();
            }
            m_eState = eStateWindowMode;
        }
    }
}

void CScreenChanger::ShowMenubar() {
    if (m_eState == eStateFullScreen && IsReadyMenubar() && !m_bShowMenu) {
        m_bShowMenu = true;
        m_rWindow.SetForeground(); // CƃANeBuɂƂB
     // m_pMenubar->SetRect(CRect(m_rWindow.GetClientSize().cx, ::GetSystemMetrics(SM_CYMENU)));
        m_pMenubar->SetRect(CRect(m_rWindow.GetClientPos(), m_rWindow.GetClientSize().cx, ::GetSystemMetrics(SM_CYMENU) - 1));
        m_pMenubar->Show();
        m_pMenubar->SetTopMost(true);
        m_pMenubar->SetForeground();
        m_pMenubar->Update();
    }
}

void CScreenChanger::HideMenubar() {
    if (IsReadyMenubar() && m_bShowMenu) {
        m_bShowMenu = false;
        m_pMenubar->SetPos(CPoint(0, - ::GetSystemMetrics(SM_CYMENU)));   //ʒuړ
        m_pMenubar->Hide();     //B
     // ǂgǔɂȂ肪ȋĈł߂Ƃ
     // if (m_rWindow.Exists()) {
     //     m_rWindow.SetForeground();
     //     m_rWindow.Update();
     // }
    }
}

bool CScreenChanger::IsReadyMenubar() {
    return (m_pMenubar->Exists() && m_pMenubar->GetMenu() != NULL);
}

void CScreenChanger::Delete() {
    //̊֐́ÃNX̃fXgN^ƁACWM_DESTROYӂ肩Ă΂B
    if (m_pMenubar->Exists()) {
        if (!m_bDeleted) {
            m_bDeleted = true;
            // dsh~B
            if (m_eState == eStateFullScreen) {
                m_eState = eStateWindowMode; // ÔߐɃZbg
             // HideMenubar(); // ǂ݂̂폜̂ŗvȂ͂
             // m_rWindow.ShowCursor(); ꍇɂĂ̓C͎ł̂ŁBB
                if (CDisplay::Restore()) {
                 // ::Sleep(250);
                    ::InvalidateRect(NULL, NULL, FALSE);
                }
            }
            m_eState = eStateWindowMode;
            m_pMenubar->Close();
        }
    }
}

#define ___LOGGING 1
#if ___LOGGING
# define ___LOG(_a) Themis_DebugLog(_a)
#else
# define ___LOG(_a) ((void)0)
#endif

bool CScreenChanger::HookMsg(CMsgHookPrms& rPrms) {
    if (m_rWindow.GetHandle() == rPrms.GetWnd()) {
        // C̃bZ[W
        if (!m_pMenubar->Exists()) {
            return false;
        } else if (!m_rWindow.Exists()) {
            Delete();
            return false;
        }
        switch (rPrms.GetMsg()) {
        case WM_SIZE:
            if (rPrms.GetWPrm() == SIZE_MINIMIZED) {
                if (m_eState == eStateFullScreen) {
                    ___LOG(L"WM_SIZE_MINIMIZED: enter");
                    // ꎞIɃtXN[
                    m_eState = eStateTempWindowMode;
                    CDisplay::Restore();
                    ___LOG(L"WM_SIZE_MINIMIZED: leave");
                    return true;
                }
            } else if (rPrms.GetWPrm() == SIZE_RESTORED) {
                if (m_eState == eStateTempWindowMode) {
                    ___LOG(L"WM_SIZE_RESTORED: enter");
                    // ANeBuɖ߂̂ŃtXN[
                    m_eState = eStateFullScreen;
                    m_rWindow.SetForeground();
                    ChangeDisplay();
                    ___LOG(L"WM_SIZE_RESTORED: leave");
                    return true;
                }
            }
            return false;

        case WM_ACTIVATEAPP:
            if (!m_rWindow.IsMinimized()) {
                if (static_cast<BOOL>(rPrms.GetWPrm()) == FALSE) {
                    if (!CWindow::IsOwnWindowThread(static_cast<DWORD>(rPrms.GetLPrm()))) {
                        if (m_eState == eStateFullScreen) {
                            ___LOG(L"WM_ACTIVATEAPP(ToWindowMode): enter");
                            if (minOnDeactive) {
                                m_rWindow.Show(SW_SHOWMINNOACTIVE); // ANeBuōŏB
                            } else {
                                // ꎞIɃtXN[
                                ToWindowMode();
                                m_eState = eStateTempWindowMode;
                            }
                            ___LOG(L"WM_ACTIVATEAPP(ToWindowMode): leave");
                        }
                    }
                } else {
                    if (m_eState == eStateTempWindowMode) {
                        ___LOG(L"WM_ACTIVATEAPP(ToFullScreen): enter");
                        // ANeBuɖ߂̂ŃtXN[
                        ToFullScreen();
                        m_eState = eStateFullScreen;
                        ___LOG(L"WM_ACTIVATEAPP(ToFullScreen): leave");
                    }
                }
            }
            return false;

        case WM_SYSCOMMAND: {
            CLocalExclusion::CLock oLock(*m_pExclusion);
            if (oLock.IsLocked()) {
                if ((rPrms.GetWPrm() & 0xFFF0) == SC_KEYMENU) {     //AltL[ł̃j[I
                    if (m_eState == eStateFullScreen) {
                        ShowMenubar();
                     // m_pMenubar->SendMsg(rPrms.GetMsg(), rPrms.GetWPrm(), rPrms.GetLPrm());
                        m_pMenubar->PostMsg(rPrms.GetMsg(), rPrms.GetWPrm(), rPrms.GetLPrm());
                        return true;
                    }
                }
            }
            }return false;

        case WM_MOUSEMOVE:
            if (rPrms.GetLPrmPoint().y <= ::GetSystemMetrics(SM_CYMENU)) {
                ShowMenubar();
            } else {
                m_pMenubar->PostMsg(WM_MOUSEMOVE, 0, 0);
            }
            return true;

        case WM_DESTROY:
            ___LOG(L"WM_DESTROY: enter");
            Delete();
            ___LOG(L"WM_DESTROY: leave");
            return true;

        default:
            return false;
        }
    } else if (rPrms.GetWnd() == m_pMenubar->GetHandle()) {
        // j[o[̃bZ[W
        switch (rPrms.GetMsg()) {
        case WM_COMMAND:
#if 0
            // ꂾƁA郁j[Ƃł̃NX̃fXgN^ĂяoƂɃfbhbNĂ܂B
            // 炾Ɖ̃ANeBu]X̖͏oȂčςނ̂BB
            m_rWindow.SendMsg(rPrms.GetMsg(), rPrms.GetWPrm(), rPrms.GetLPrm());
#else
            // j[IĂɗ͂Ȃ̂ŁAƉBĂ܂B
            // ĂƁAt@CI_CAOƂoꍇɃANeBuɂȂȂB
            // cǂΏÖ@BB(L`)
            HideMenubar();
            m_rWindow.PostMsg(rPrms.GetMsg(), rPrms.GetWPrm(), rPrms.GetLPrm());
#endif
            return true;

        case WM_MOUSEMOVE:
            if (::GetSystemMetrics(SM_CYMENU) < m_rWindow.ToClient(CSystem::GetCursorPos()).y) {
                HideMenubar();
            }
            return true;

        case WM_ACTIVATEAPP:
            if (static_cast<BOOL>(rPrms.GetWPrm()) == FALSE) {
                // tH[JXۂ̂ŉB
                HideMenubar();
            } else if (m_eState != eStateFullScreen) {
                // Ô߁BB
                HideMenubar();
            }
            return true;

        case WM_CLOSE:
            m_pMenubar->Delete();
            return true;

        default:
            return false;
        }
    } else {
        return false;
    }
}

}
