#pragma once
#include "File.h"
#include "String/Encoding.h"
namespace Themis {

/**
 * ǂݍ݃f[^Xg[B
 *
 * ȊŎgׂB
 * @code
 * do {
 *     Func(rSrcStream.GetBuffer(), rSrcStream.GetBufferSize());
 * } while (rSrcStream.ToNextBuffer()); // falseŏI
 * @endcode
 */
class IReadStream : private boost::noncopyable {
public:
    virtual ~IReadStream() {}
    virtual const void* GetBuffer() const = 0;
    virtual size_t GetBufferSize() const = 0;
    virtual bool ToNextBuffer() const = 0;
    /// ŏɃg[^TCYȂƕsւȏꍇ̂ŁBB
    virtual size_t GetTotalSize() const = 0;
};

/**
 * void*  std::string  std::vector<uint8_t> B
 *
 * g̓RsȂ̂ŁAIuWFNg̎ɒӂׂB
 */
class CReadStream : public IReadStream {
    const void* memory;
    size_t size;
public:
    CReadStream(void* p, size_t n) : memory(p), size(n) {}
    CReadStream(const vector<uint8_t>& o) : memory(&o[0]), size(o.size()) {}
    const void* GetBuffer() const { return memory; }
    size_t GetBufferSize() const { return size; }
    bool ToNextBuffer() const { return false; }
    size_t GetTotalSize() const { return size; }
};

/**
 * ݃f[^Xg[B
 * ǂݍ݂ɔׂ犄ƒPB
 */
class IWriteStream : private boost::noncopyable {
public:
    virtual ~IWriteStream() {}
    virtual void Write(void* p, size_t n) = 0;
};

/**
 * std::vector<uint8_t>Ƀf[^~WriteStream
 */
class CWriteStream : public IWriteStream {
    vector<uint8_t>* buffer;
    bool owner;
public:
    CWriteStream() : buffer(new vector<uint8_t>), owner(true) {}
    CWriteStream(vector<uint8_t>* p) : buffer(p), owner(false) {}
    ~CWriteStream() { if (owner) delete buffer; }
    void Write(void* p, size_t n) {
        buffer->insert(buffer->end(), static_cast<const uint8_t*>(p), static_cast<const uint8_t*>(p) + n);
    }
    const vector<uint8_t>& GetBuffer() { return *buffer; }
};

/**
 * t@CStream
 */
class CFileWriteStream : public IWriteStream {
    auto_ptr<IFile> file;
public:
    CFileWriteStream(const CFileName& oFileName) {
        CFile(oFileName).CreateDir();
        file.reset(new CStdFile(oFileName, L"wb"));
    }
    ~CFileWriteStream() {}
    void Write(void* p, size_t n){ file->Write(p, n); }
    IFile& GetFile() { return *file; }
};

/**
 * IWriteStreamւ݂̏⏕NX
 */
class CStreamWriter : private boost::noncopyable {
    IWriteStream& stream;
    size_t pos;
    uint8_t buffer[0x200];
public:
    inline CStreamWriter(IWriteStream& r) : stream(r), pos(0) {}
    inline ~CStreamWriter() { if (0 < pos) { Flush(); } }

    inline void Write(char t) { Write(&t, sizeof(t)); }
    inline void Write(short t) { Write(&t, sizeof(t)); }
    inline void Write(int t) { Write(&t, sizeof(t)); }
    inline void Write(long t) { Write(&t, sizeof(t)); }
    inline void Write(unsigned char t) { Write(&t, sizeof(t)); }
    inline void Write(unsigned short t) { Write(&t, sizeof(t)); }
    inline void Write(unsigned int t) { Write(&t, sizeof(t)); }
    inline void Write(unsigned long t) { Write(&t, sizeof(t)); }
    inline void Write(const vector<uint8_t>& t) { if (!t.empty()) Write(&t[0], t.size()); }
    inline void Write(const void* p, size_t n) {
        size_t t = pos + n;
        if (sizeof(buffer) < t) {
            Flush();
            memcpy(buffer, p, n);
            pos = n;
        } else {
            memcpy(buffer + pos, p, n);
            pos = t;
        }
    }

private:
    inline void Flush() {
        stream.Write(buffer, pos);
        pos = 0;
    }
};

}
