#include "stdafx.h"
#include "WorkerThread.h"
#include "ThreadedWork.h"
#include "Thread.h"
#include "../ErrorHandler.h"
namespace Themis {

WorkerThread::WorkerThread() {
    work.reset(new ThreadedWork());
    threadValid = true;
    thread.reset(new boost::thread(std::tr1::bind(&WorkerThread::OnThread, this)));
}

WorkerThread::WorkerThread(const std::tr1::function<void()>& w) {
    work.reset(new ThreadedWork(w));
    threadValid = true;
    thread.reset(new boost::thread(std::tr1::bind(&WorkerThread::OnThread, this)));
}

WorkerThread::~WorkerThread() {
    {
        boost::mutex::scoped_lock lock(mutex);
        threadValid = false;
        condition.notify_all();
    }
    thread->join();
    Themis_DebugAssert(work->IsFinished());
}

/// Vds
void WorkerThread::SetWork(const std::tr1::function<void()>& w) {
    Themis_DebugAssert(threadValid);
    work->Join();
    work->Reset(w);
    boost::mutex::scoped_lock lock(mutex);
    Themis_DebugAssert(!work->IsFinished());
    condition.notify_all();
}

/// sȂtrue
bool WorkerThread::IsIdling() const {
    return work->IsFinished();
}

/// workI܂ő҂B
void WorkerThread::Join() {
    work->Join();
}

/// XbhB
void WorkerThread::OnThread() {
    Thread::Initialize();
    while (threadValid) {
        if (work->IsFinished()) {
            // VdZbg܂ő҂B
            boost::mutex::scoped_lock lock(mutex);
            if (threadValid && work->IsFinished()) { // ̃`FbNKvH
                Thread::SetName(L"WorkerThread(ҋ@)");
                condition.wait(lock);
                Themis_DebugAssert(!threadValid || !work->IsFinished());
            }
        } else {
            // d̎s
            try {
                work->DoWork();
            } catch (...) {
                Themis_DebugWarning(L"WorkerThreadŗO");
                // O͈ԂBǂōđĂboost::threadcatchႤ̂ŁBB
            }
        }
    }
    Thread::SetName(L"WorkerThread(I)"); // ǁ[̏ԂŃnO鎖悤ȁBB
}

#ifndef NDEBUG

namespace {
    static void SetThreadName(const std::tr1::function<void()>& w, const string& threadName) {
        Thread::SetName(threadName);
        w();
    }
}

/// Xbhݒ肷邨dԂ
std::tr1::function<void()> WorkerThread::GetWorkWrapper(const std::tr1::function<void()>& w, const string& threadName) {
    return std::tr1::bind(&SetThreadName, w, threadName);
}

#endif

}
