#include "stdafx.h"
#include "TimeDate.h"
#include "Utility/StringUtility.h"
#include <time.h>
namespace Themis {

CTimeDate::CTimeDate() : m_nYear(INT_MIN), m_nDate(INT_MIN), m_nTime(INT_MIN) {
}

CTimeDate::CTimeDate(int nYear, int nDate, int nTime) : m_nYear(nYear), m_nDate(nDate), m_nTime(nTime) {
}

CTimeDate::~CTimeDate() {
}

bool CTimeDate::IsValid() const {
    return m_nYear != INT_MIN && m_nDate != INT_MIN && m_nTime != INT_MIN;
}

namespace {
    // [NȂ̂ǂ
    inline int isLeapYear(int year) {
        return ((year % 4 == 0) && (year % 100 != 0)) || (year % 400 == 0);
    }
    // w茎̓擾
    inline int getMonthDays(int year, int month) {
        if (month == 2) {
            return isLeapYear(year) ? 29 : 28;
        } else {
            const int d[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
            return d[month - 1];
        }
    }
    // j̎擾
    inline int getWDay(int year, int month, int day) {
        // 2005/01/01 == yj ƂĎZo
        int wday = 6; //  yj
        for (int i = 2005; i < year; i++) {
            wday += isLeapYear(i) ? 366 : 365;
        }
        for (int i = 2005; year < i; i--) {
            wday -= isLeapYear(i - 1) ? 366 : 365;
            wday += 371; // 7*53BlɂƖʓ|Ȃ̂ŁBB
        }
        for (int i = 1; i < month; i++) {
            wday += getMonthDays(year, i);
        }
        wday += day - 1;
        return wday % 7;
    }
}

/// N̉Z
void CTimeDate::AddYear(int n) {
    m_nYear += n;
}

/// t̉Z
void CTimeDate::AddDate(int n) {
    m_nDate += n;
    int nYDay;
    while ((nYDay = isLeapYear(GetYear()) ? 366 : 365) <= m_nDate) { // Jオ
        m_nDate -= nYDay;
        AddYear(1);
    }
}

/// Ԃ̉Z
void CTimeDate::AddTime(int n) {
    m_nTime += n;
    while (24 * 3600 <= m_nTime) { // Jオ
        m_nTime -= 24 * 3600;
        AddDate(1);
    }
}

/// ŇZ
void CTimeDate::SubYear(int n) {
    m_nYear -= n;
}

/// ťZ
void CTimeDate::SubDate(int n) {
    m_nDate -= n;
    while (m_nDate < 0) { // J艺
        SubYear(1);
        m_nDate += isLeapYear(GetYear()) ? 366 : 365;
    }
}

/// Ԃ̌Z
void CTimeDate::SubTime(int n) {
    m_nTime -= n;
    while (m_nTime < 0) { // J艺
        m_nTime += 24 * 3600;
        SubDate(1);
    }
}

/// ZB
CTimeDate& CTimeDate::operator-=(const CTimeDate& r) {
    Themis_DebugAssert(IsValid());
    Themis_DebugAssert(r.IsValid());
    SubTime(r.GetTime());
    SubDate(r.GetDate());
    SubYear(r.GetYear());
    Themis_DebugAssert(0 <= GetYear());
    return *this;
}

/// ZB
CTimeDate& CTimeDate::operator+=(const CTimeDate& r) {
    Themis_DebugAssert(IsValid());
    Themis_DebugAssert(r.IsValid());
    AddTime(r.GetTime());
    AddDate(r.GetDate());
    AddYear(r.GetYear());
    return *this;
}

/// b֕ϊ
int CTimeDate::ToSec() const {
    Themis_DebugAssert(m_nYear == 0);
    return m_nDate * 24 * 3600 + m_nTime;
}

/// j̎擾(0-6)
int CTimeDate::GetWDay() const {
    Themis_DebugAssert(IsValid());
    const int year = GetYear();
    int date = GetDate() + 1;
    for (int i = 1; i <= 12; i++) {
        const int t = getMonthDays(year, i);
        if (date <= t) {
            return getWDay(year, i, date);
        }
        date -= t;
    }
    throw RuntimeException(L"j̎擾Ɏs܂");
}

/// ̎擾(1-12)
int CTimeDate::GetMonth() const {
    Themis_DebugAssert(IsValid());
    const int year = GetYear();
    int date = GetDate() + 1;
    for (int i = 1; i <= 12; i++) {
        const int t = getMonthDays(year, i);
        if (date <= t) {
            return i;
        }
        date -= t;
    }
    throw RuntimeException(L"̎擾Ɏs܂");
}

/// t̎擾(1-31)
int CTimeDate::GetDay() const {
    Themis_DebugAssert(IsValid());
    const int year = GetYear();
    int date = GetDate() + 1;
    for (int i = 1; i <= 12; i++) {
        int t = getMonthDays(year, i);
        if (date <= t) {
            return date;
        }
        date -= t;
    }
    throw RuntimeException(L"t̎擾Ɏs܂");
}

string CTimeDate::ToStr() const {
    if (!IsValid()) {
        return string();
    }
    bool bSwitch = false;
    string str;
    if (0 < GetYear()) {
        bSwitch = true;
        str += ItoA(GetYear());
        str += L"N";
    }
    if (bSwitch || 0 < GetDate()) {
        bSwitch = true;
        str += ItoA(GetDate());
        str += L"";
    }
    if (bSwitch || 0 < GetTime() / 3600) {
        bSwitch = true;
        str += ItoA(GetTime() / 3600);
        str += L""; // L""ȂBB
    }
    if (bSwitch || 0 < (GetTime() / 60) % 60) {
        bSwitch = true;
        str += ItoA((GetTime() / 60) % 60);
        str += L"";
    }
    if (bSwitch || 0 < GetTime() % 60) {
        bSwitch = true;
        str += ItoA(GetTime() % 60);
        str += L"b";
    }
    return str;
}

CTimeDate CTimeDate::GetNow() {
    time_t tt = time(NULL);
    tm* t = localtime(&tt);
    return CTimeDate(t->tm_year + 1900, t->tm_yday, t->tm_hour * 3600 + t->tm_min * 60 + t->tm_sec);
}

}
