#include "stdafx.h"
#include <math.h>
#include "Windows/System.h"
#include "Wnd.h"
#include "MsgHook.h"
#include "WindowSnapper.h"
namespace Themis {

CWindowSnapper::CWindowSnapper() : m_bSnapWorkArea(true), m_nSnapDist(32) {
}

CWindowSnapper::~CWindowSnapper() {
    Themis_DebugAssert(m_oWindowToCursorList.empty());
}

bool CWindowSnapper::IsSnapWorkArea() const {
    return m_bSnapWorkArea;
}

void CWindowSnapper::SetSnapWorkArea(bool b) {
    m_bSnapWorkArea = b;
}

int CWindowSnapper::GetSnapDist() const {
    return m_nSnapDist;
}

void CWindowSnapper::SetSnapDist(int n) {
    m_nSnapDist = n;
}

void CWindowSnapper::AddTarget(HWND hWnd) {
    if (std::find(m_vctTarget.begin(), m_vctTarget.end(), hWnd) == m_vctTarget.end()) {
        m_vctTarget.push_back(hWnd); // d͔F߂Ȃ
    }
}

void CWindowSnapper::RemoveTarget(HWND hWnd) {
    vector<HWND>::iterator it = std::find(m_vctTarget.begin(), m_vctTarget.end(), hWnd);
    if (it != m_vctTarget.end()) {
        m_vctTarget.erase(it);
    }
}

void CWindowSnapper::ClearTarget() {
    m_vctTarget.clear();
}

/// WM_ENTERSIZEMOVE ̏
bool CWindowSnapper::OnEnterSizeMove(CMsgHookPrms& rPrms) {
    OnBeginMove(rPrms.GetWnd());
    return true; // falseɂƂׂcH
}

/// WM_EXITSIZEMOVE ̏
bool CWindowSnapper::OnExitSizeMove(CMsgHookPrms& rPrms) {
    OnEndMove(rPrms.GetWnd());
    return true; // falseɂƂׂcH
}

/// WM_MOVING̏
bool CWindowSnapper::OnMoving(CMsgHookPrms& rPrms) {
    Themis_DebugAssert(m_oWindowToCursorList.find(rPrms.GetWnd()) != m_oWindowToCursorList.end());
    map_vector<HWND, CSize>::iterator it = m_oWindowToCursorList.find(rPrms.GetWnd());
    if (it != m_oWindowToCursorList.end()) {
        CRect rc(CSystem::GetCursorPos() - (*it).second, CWnd(rPrms.GetWnd()).GetSize());
        if (OnMove(rPrms.GetWnd(), rc)) {
            *(RECT*)rPrms.GetLPrm() = rc;
            return true;
        }
    }
    return false;
}

void CWindowSnapper::OnBeginMove(HWND hWnd) {
    m_oWindowToCursorList[hWnd] = CSystem::GetCursorPos() - CWnd(hWnd).GetPos();
}

void CWindowSnapper::OnEndMove(HWND hWnd) {
    m_oWindowToCursorList.erase(hWnd);
}

namespace {
    bool IsIntersectV(const CRect& rc1, const CRect& rc2) {
        return rc2.top <= rc1.bottom && rc1.top <= rc2.bottom;
    }
    bool IsIntersectH(const CRect& rc1, const CRect& rc2) {
        return rc2.left <= rc1.right && rc1.left <= rc2.right;
    }
    bool Snapping(int nDist, long& a, long& b, long& c, long& d, int nSnapDist) {
        if (abs(nDist) <= nSnapDist) {
            a += nDist;
            b += nDist;
            c += nDist;
            d += nDist;
            return true;
        }
        return false;
    }
    // ʒ[ւ̃Xibv
    bool SnappingInside(CRect& rrc, CRect& rcWindow, const CRect& rcArea, int nSnapDist) {
        bool bSnap = false;
        if (IsIntersectV(rcWindow, rcArea) && Snapping(rcArea.left - rcWindow.left, rrc.left, rrc.right, rcWindow.left, rcWindow.right, nSnapDist)) { // 
            bSnap = true;
        }
        if (IsIntersectH(rcWindow, rcArea) && Snapping(rcArea.top - rcWindow.top, rrc.top, rrc.bottom, rcWindow.top, rcWindow.bottom, nSnapDist)) { // 㑤
            bSnap = true;
        }
        if (IsIntersectV(rcWindow, rcArea) && Snapping(rcArea.right - rcWindow.right, rrc.left, rrc.right, rcWindow.left, rcWindow.right, nSnapDist)) { // E
            bSnap = true;
        }
        if (IsIntersectH(rcWindow, rcArea) && Snapping(rcArea.bottom - rcWindow.bottom, rrc.top, rrc.bottom, rcWindow.top, rcWindow.bottom, nSnapDist)) { // 
            bSnap = true;
        }
        return bSnap;
    }
    // ̃EBhEւ̃Xibv
    bool SnappingOutside(CRect& rrc, CRect& rcWindow, const CRect& rcArea, int nSnapDist) {
        bool bSnap = false;
        if (IsIntersectV(rcWindow, rcArea) && Snapping(rcArea.right - rcWindow.left, rrc.left, rrc.right, rcWindow.left, rcWindow.right, nSnapDist)) { // 
            bSnap = true;
        }
        if (IsIntersectH(rcWindow, rcArea) && Snapping(rcArea.bottom - rcWindow.top, rrc.top, rrc.bottom, rcWindow.top, rcWindow.bottom, nSnapDist)) { // 㑤
            bSnap = true;
        }
        if (IsIntersectV(rcWindow, rcArea) && Snapping(rcArea.left - rcWindow.right, rrc.left, rrc.right, rcWindow.left, rcWindow.right, nSnapDist)) { // E
            bSnap = true;
        }
        if (IsIntersectH(rcWindow, rcArea) && Snapping(rcArea.top - rcWindow.bottom, rrc.top, rrc.bottom, rcWindow.top, rcWindow.bottom, nSnapDist)) { // 
            bSnap = true;
        }
        return bSnap;
    }
}

bool CWindowSnapper::OnMove(HWND hWnd, CRect& rrc) {
    map_vector<HWND, CSize>::iterator it = m_oWindowToCursorList.find(hWnd);
    if (it == m_oWindowToCursorList.end()) {
        Themis_DebugWarning(L"XibvɎs");
    } else {
        CWnd oWnd(hWnd);
        CRect rcWindow(CSystem::GetCursorPos() - it->second, rrc.GetSize()); // ړ\ʒu
        bool bSnap = false;
        for (size_t i = 0, n = m_vctTarget.size(); i < n; i++) { // ̃EBhEւ̃Xibv
            if (hWnd != m_vctTarget[i]) {
                if (SnappingOutside(rrc, rcWindow, CWnd(m_vctTarget[i]).GetRect(), m_nSnapDist)) {
                    bSnap = true;
                }
            }
        }
        if (m_bSnapWorkArea) { // ʒ[ւ̃Xibv
            vector<CRect> vctWorkAreas = CSystem::GetAllWorkAreas();
            for (size_t i = 0, n = vctWorkAreas.size(); i < n; i++) {
                if (SnappingInside(rrc, rcWindow, vctWorkAreas[i], m_nSnapDist)) {
                    bSnap = true;
                }
            }
        }
        return bSnap;
    }
    return false;
}

/// XibvTCY
void CWindowSnapper::SetWindowRect(HWND hWnd, const CRect& rc) {
    CWnd oWindow(hWnd);
    oWindow.SetRect(ResizeRect(hWnd, oWindow.GetRect(), rc));
}

/// XibvTCY
void CWindowSnapper::SetClientRect(HWND hWnd, const CRect& rc) {
    CWnd oWindow(hWnd);
    oWindow.SetClientRect(ResizeRect(hWnd, oWindow.GetClientRect(), rc));
}

namespace {
    // ̃EBhEɃXibvĂ邩ǂ
    void GetSnapDistOutside(const CRect& rcWindow, const CRect& rcArea, bool& bR, bool& bB, int nSnapDist) {
        if (IsIntersectV(rcWindow, rcArea) && abs(rcArea.left - rcWindow.right) <= nSnapDist) { // E
            bR = true;
        }
        if (IsIntersectH(rcWindow, rcArea) && abs(rcArea.top - rcWindow.bottom) <= nSnapDist) { // 
            bB = true;
        }
    }
    // ʒ[ɃXibvĂ邩ǂ
    void GetSnapDistInside(const CRect& rcWindow, const CRect& rcArea, bool& bR, bool& bB, int nSnapDist) {
        // ʒ[DȂ̂ŁA[Ə[͑̃EBhEւ̃Xibvł
        if (IsIntersectV(rcWindow, rcArea) && abs(rcArea.left - rcWindow.left) <= nSnapDist) { // 
            bR = false;
        }
        if (IsIntersectH(rcWindow, rcArea) && abs(rcArea.top - rcWindow.top) <= nSnapDist) { // 㑤
            bB = false;
        }
        if (IsIntersectV(rcWindow, rcArea) && abs(rcArea.right - rcWindow.right) <= nSnapDist) { // E
            bR = true;
        }
        if (IsIntersectH(rcWindow, rcArea) && abs(rcArea.bottom - rcWindow.bottom) <= nSnapDist) { // 
            bB = true;
        }
    }
}

/// ƂɃXibvĂꍇŒ肵ăTCY`Ԃ
CRect CWindowSnapper::ResizeRect(HWND hWnd, const CRect& rcWindow, const CRect& rc) {
    bool bR = false, bB = false;
    for (size_t i = 0, n = m_vctTarget.size(); i < n; i++) { // ̃EBhEւ̃Xibv
        if (hWnd != m_vctTarget[i]) {
            GetSnapDistOutside(rcWindow, CWnd(m_vctTarget[i]).GetRect(), bR, bB, m_nSnapDist);
        }
    }
    if (m_bSnapWorkArea) {
        vector<CRect> vctWorkAreas = CSystem::GetAllWorkAreas();
        for (size_t i = 0, n = vctWorkAreas.size(); i < n; i++) {
            GetSnapDistInside(rcWindow, vctWorkAreas[i], bR, bB, m_nSnapDist);
        }
    }
    CRect rc2 = rc;
    if (bR) {
        rc2.left = rcWindow.right - rc.GetWidth();
        rc2.right = rcWindow.right;
    }
    if (bB) {
        rc2.top = rcWindow.bottom - rc.GetHeight();
        rc2.bottom = rcWindow.bottom;
    }
    return rc2;
}

}
