
#include "stdafx.h"
#include <Themis/Windows/System.h>
#include <Themis/Windows/Process.h>
#include <Themis/AppFrameBase.h>
#include <Themis/Allocator.h>
#include <Themis/Window.h>
#include <Themis/TrayIcon.h>
#include <Themis/Graphics.h>
#include <Themis/Image.h>
#include <Themis/Config.h>
#include <Themis/File.h>
#include <Themis/KeyTable.h>
#include <Themis/BatchFile.h>
#include <Themis/using.h>
#include "../../Hook/Hook/Hook.h"
#include <mmsystem.h>
#pragma comment(lib, "winmm.lib")
#include "Dialogs.h"
#include "resource.h"


const int c_nBuildCount = 19;


bool g_bReboot = false;

class CAppWindow: private IMsgHooker {
    auto_ptr<CWindow>       m_pWindow;
    auto_ptr<CTrayIcon>     m_pTrayIcon;
    CConfig                 m_oConfig;
    HWND                    m_hNextWindow;
    bool                    m_bWatchClipboard;
    pair<UINT, int>         m_oHotKey[3];
public:
    CAppWindow() {
        m_pWindow.reset(new CWindow);
        m_pWindow->GetHook().Add(this);

        m_oConfig.Bool(L"CaptureCursor", false);
        m_oConfig.Bool(L"CaptureLayer", false);
        m_oConfig.Num(L"DoubleClick", 0);
        m_oConfig.File(L"FileNameFormat", L"ScreenShot/ss%03d.png");
        m_oConfig.Bool(L"KeyHook_", false);
        m_oConfig.Indirect(L"KeyHook0", L"Win+P");
        m_oConfig.Indirect(L"KeyHook1", L"Alt+Win+P");
        m_oConfig.Indirect(L"KeyHook2", L"Shift+Win+P");
        m_oConfig.Num(L"SaveType", 0);
        m_oConfig.Num(L"JpgQuality(%)", 80);
        m_oConfig.Num(L"PngCompress", 6);
        m_oConfig.Bool(L"WatchClipboard", true);
        m_oConfig.Bool(L"NotifyBeep", false);
        m_oConfig.Bool(L"NotifyWav", false);
        m_oConfig.File(L"NotifyWavFile", L"%windir%\\Media\\chimes.wav");
        m_oConfig.Load();
        m_oConfig.Delete(L"KeyHook");
        m_oConfig.Save();

        m_bWatchClipboard = m_oConfig.Bool(L"WatchClipboard");
        if (m_oConfig.Bool(L"KeyHook_")) {
            m_oHotKey[0] = CKeyTable::GetVKey(m_oConfig.Indirect(L"KeyHook0"));
            m_oHotKey[1] = CKeyTable::GetVKey(m_oConfig.Indirect(L"KeyHook1"));
            m_oHotKey[2] = CKeyTable::GetVKey(m_oConfig.Indirect(L"KeyHook2"));
            SetKeyHook(&CAppWindow::KeyHook, this);
        }
    }

    ~CAppWindow() {
        UnsetKeyHook();
    }

    int Run(CAppFrameBase& rAppFrame) {
        CWindowParameter oPrm;
        oPrm.GetStyle() = WS_POPUP;
        oPrm.GetExStyle() = WS_EX_TOOLWINDOW;
        m_pWindow->Create(oPrm);
        m_pWindow->SetRect(CRect(-1, -1, 0, 0));
        m_pWindow->WaitForQuit();
        return 0;
    }

private:
    static bool WINAPI KeyHook(INT nCode, WPARAM wParam, LPARAM lParam, void* pParam) {
        KBDLLHOOKSTRUCT* pkbhs = (KBDLLHOOKSTRUCT*)lParam;
        if (nCode == HC_ACTION && (pkbhs->flags & LLKHF_UP) == 0) {
            CAppWindow* pThis = (CAppWindow*)pParam;
            try {
                int nModifiers = CKeyTable::GetModifiers();
                if (pThis->m_oHotKey[0].first == pkbhs->vkCode && pThis->m_oHotKey[0].second == nModifiers) {
                    pThis->CaptureScreen();
                    return true;
                } else if (pThis->m_oHotKey[1].first == pkbhs->vkCode && pThis->m_oHotKey[1].second == nModifiers) {
                    pThis->CaptureActiveWindow();
                    return true;
                } else if (pThis->m_oHotKey[2].first == pkbhs->vkCode && pThis->m_oHotKey[2].second == nModifiers) {
                    pThis->CaptureActiveWindowClient();
                    return true;
                }
            } catch (std::exception& e) {
                pThis->m_pWindow->MsgBox(L"G[", e);
            }
        }
        return false;
    }
    bool HookMsg(CMsgHookPrms& rPrms) {
        if (rPrms.GetMsg() == WM_COMMAND) {
            switch (rPrms.GetWPrmLoWord()) {
            case ID__40002: //摜̃tH_J
                try {
                    CFileName oFileName(m_oConfig.File(L"FileNameFormat"), CSystem::GetDesktopDir());
                    Process::Start(oFileName.GetDir());
                } catch (std::exception&) {
                    // Ƃ肠
                }
                break;
            case ID__40003: // ݒ
                if (CConfigDialog(m_oConfig).Open(m_pWindow->GetHandle()) == DialogResult::OK) {
                    m_oConfig.Save();
                    // ݒLɂ邽߂ɍċN
                    g_bReboot = true;
                    m_pWindow->Close();
                }
                break;
            case ID__40004: // 蓮Lv`[
                CaptureScreen();
                break;
            case ID__40005: // st@ĈtH_J
                Process::Start(CFileNameModuleDir().Get());
                break;
            case ID__40006: // zzTCgJ
                Process::Start(L"http://hp.vector.co.jp/authors/VA039571/");
                break;
            case ID__40008: // ҂փ[𑗂
                Process::Start(L"http://tqzh.tk:24497/site/mail/");
                break;
            case ID__40007: // ċN
                g_bReboot = true;
                m_pWindow->Close();
                break;
            case ID__40001: //I
                m_pWindow->Close();
                break;
            default:
                return false;
            }
            return true;
        } else if (rPrms.GetMsg() == WM_APP + 1) { // }EXbZ[W
            UINT nMouseMsg = m_pTrayIcon->GetMouseMsg(rPrms);
            if (nMouseMsg == WM_RBUTTONUP) {
                m_pWindow->ShowPopupMenu(IDR_MENU1);
                return true;
            } else if (nMouseMsg == WM_LBUTTONDBLCLK) {
                switch (m_oConfig.Num(L"DoubleClick")) {
                case 0: m_pWindow->PostMsg(WM_COMMAND, ID__40002, 0); break; //摜̃tH_J
                case 1: m_pWindow->PostMsg(WM_COMMAND, ID__40003, 0); break; //RtBOt@CJ
                case 2: m_pWindow->PostMsg(WM_COMMAND, ID__40004, 0); break; //蓮Lv`[
                }
                return true;
            } else {
                return false;
            }
        } else if (m_pTrayIcon.get() != NULL && m_pTrayIcon->OnTaskbarCreate(rPrms)) { // ACRĕ\
            return true;
        } else if (rPrms.GetMsg() == WM_DRAWCLIPBOARD) {
            CaptureClipboard();
            if (m_hNextWindow != NULL) {
                ::SendMessage(m_hNextWindow , rPrms.GetMsg() , rPrms.GetWPrm() , rPrms.GetLPrm());
            }
            return true;
        } else if (rPrms.GetMsg() == WM_CHANGECBCHAIN) {
            if ((HWND)rPrms.GetWPrm() == m_hNextWindow) {
                m_hNextWindow = (HWND)rPrms.GetLPrm();
            } else if (m_hNextWindow != NULL) {
                ::SendMessage(m_hNextWindow , rPrms.GetMsg() , rPrms.GetWPrm() , rPrms.GetLPrm());
            }
            return true;
        } else if (rPrms.GetMsg() == WM_CREATE) {
            if (m_bWatchClipboard) {
                m_hNextWindow = ::SetClipboardViewer(m_pWindow->GetHandle());
            } else {
                m_hNextWindow = NULL;
            }
            m_pTrayIcon.reset(new CTrayIcon(m_pWindow->GetHandle(), WM_APP + 1));
            m_pTrayIcon->SetIcon(IDI_ICON1);
            m_pTrayIcon->SetTip(string::Format(L"Ղ肷 build-%d", c_nBuildCount));
            return true;
        } else if (rPrms.GetMsg() == WM_CLOSE) {
            m_pWindow->Delete();
            return true;
        } else if (rPrms.GetMsg() == WM_DESTROY) {
            if (m_bWatchClipboard) {
                ::ChangeClipboardChain(m_pWindow->GetHandle() , m_hNextWindow);
            }
            m_pTrayIcon->UnsetIcon();
            return true;
        }
        return false;
    }

    void CaptureClipboard() {
        CDib oTempPlane(24);
        oTempPlane.SetSize(CSize(::GetSystemMetrics(SM_CXSCREEN), ::GetSystemMetrics(SM_CYSCREEN))); //Ƃ肠쐬
        CSize szClip;
        if (CSystem::GetClipboard(CPlane::CGdiLock(oTempPlane), CPoint(), szClip)) {
            CSize szPlane = oTempPlane.GetSize();
            if (szPlane.cx < szClip.cx || szPlane.cy < szClip.cy) {
                oTempPlane.SetSize(szPlane.GetUnion(szClip));
                if (!CSystem::GetClipboard(CPlane::CGdiLock(oTempPlane), CPoint(), szClip)) {
                    throw RuntimeException(L"XN[Vbg̎擾Ɏs܂");
                }
            }
            SavePlane(oTempPlane, szClip);
        }
    }
    void CaptureScreen() {
        HDC hdc = ::GetDC(NULL);
        CaptureScreen(hdc, CSystem::GetScreensCoverRect());
        ::ReleaseDC(NULL, hdc);
    }
    void CaptureActiveWindow() {
        CWnd oWnd(::GetForegroundWindow());
        CRect rc = oWnd.GetRectScreen();
        if (!oWnd.Exists()) { throw RuntimeException(L"ANeBuȃEBhE݂܂"); }
        if (!oWnd.IsVisible()) { throw RuntimeException(L"ANeBuȃEBhEsԂł"); }
        if (!rc.IsValid()) { throw RuntimeException(L"ANeBuȃEBhȆ傫ł"); }
        HDC hdc = ::GetDC(NULL);
        CaptureScreen(hdc, rc);
        ::ReleaseDC(NULL, hdc);
    }
    void CaptureActiveWindowClient() {
        CWnd oWnd(::GetForegroundWindow());
        CRect rc = oWnd.GetClientRect();
        if (!oWnd.Exists()) { throw RuntimeException(L"ANeBuȃEBhE݂܂"); }
        if (!oWnd.IsVisible()) { throw RuntimeException(L"ANeBuȃEBhEsԂł"); }
        if (!rc.IsValid()) { throw RuntimeException(L"ANeBuȃEBhȆ傫ł"); }
        HDC hdc = ::GetDC(NULL);
        CaptureScreen(hdc, rc);
        ::ReleaseDC(NULL, hdc);
    }
    void CaptureScreen(HDC hdc, const CRect& rc) {
        CDib oTempPlane(24);
        oTempPlane.SetSize(rc.GetSize());
        if (m_oConfig.Bool(L"CaptureLayer")) {
            CDC(CPlane::CGdiLock(oTempPlane)).CaptureBlt(hdc, CPoint(), rc);
        } else {
            CDC(CPlane::CGdiLock(oTempPlane)).Blt(hdc, CPoint(), rc);
        }
        if (m_oConfig.Bool(L"CaptureCursor")) { // ACR̕`
            CURSORINFO ci = {sizeof(CURSORINFO)};
            if (::GetCursorInfo(&ci) != FALSE) {
                ICONINFO ii;
                if (::GetIconInfo(ci.hCursor, &ii) != FALSE) {
                    ::DrawIcon(CPlane::CGdiLock(oTempPlane).Get(),
                        ci.ptScreenPos.x - rc.left - ii.xHotspot,
                        ci.ptScreenPos.y - rc.top - ii.yHotspot, ci.hCursor);
                }
            }
        }
        SavePlane(oTempPlane, oTempPlane.GetSize());
    }
    void SavePlane(CDib& r, const CSize& sz) {
        switch (m_oConfig.Num(L"SaveType")) {
        case 0:
            {
                CPngSaver oSaver;
                oSaver.SetCompressLevel(m_oConfig.Num(L"PngCompress"));
                oSaver.Save(MakeFileName(), r, sz.ToRect());
            }
            break;
        case 1:
            {
                CJpgSaver oSaver;
                oSaver.SetQuality(m_oConfig.Num(L"JpgQuality(%)"));
                oSaver.Save(MakeFileName(), r, sz.ToRect());
            }
            break;
        case 2:
            {
                CBmpSaver oSaver;
                oSaver.Save(MakeFileName(), r, sz.ToRect());
            }
            break;
        default:
            throw RuntimeException(L"RtBOt@C 'SaveType' ̒lsł");
        }

        if (m_oConfig.Bool(L"NotifyBeep")) {
            ::Beep(528 * 2, 100);

        }
        if (m_oConfig.Bool(L"NotifyWav")) {
            ::PlaySound(Encoding::GetDefault().GetBytes(CFileNameModuleDir(m_oConfig.File(L"NotifyWavFile")).Get()).c_str(),
                NULL, SND_FILENAME | SND_ASYNC);
        }
    };

    CFileName MakeFileName() {
        string str = m_oConfig.File(L"FileNameFormat");
        CFileName oDesktopDir(CSystem::GetDesktopDir());
        for (int i = 0; i < 0x10000; i++) {
            CFileName oFileName(string::Format(str.ToCharPtr(), i), oDesktopDir);
            if (!CFile(oFileName).Exists()) {
                // ݂Ȃ΃\Ԃ
                return oFileName;
            }
        }
        throw RuntimeException(L"t@C̍쐬Ɏs܂BKɐƂĂB");
    }
};

class CAppFrame : public CSingleAppFrameBase {
    int InnerRun() {
        return IsFirstRun() ? CAppWindow().Run(*this) : 0;
    }
};

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPTSTR lpCmdLine, int nCmdShow) {
    int nRet;
    {
        CAppFrame theApp;
        theApp.Initialize(hInstance, hPrevInstance, lpCmdLine, nCmdShow);
        nRet = theApp.Run();
    }
    if (g_bReboot) {
        CBatchFile::RunReboot(CFileNameModuleDir(L"$ps#restart$.bat"));
    }
    return nRet;
}

